"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define auxiliary functions for plots.
"""

__all__ = ['geo_bin_from_array',
           'geo_im_from_array',
           'make_map',
           'add_shapes',
           'add_populated_places',
           'add_cntry_names'
          ]

import logging
from textwrap import wrap

from scipy.interpolate import griddata
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from mpl_toolkits.axes_grid1 import make_axes_locatable
from shapely.geometry import box
import cartopy.crs as ccrs
from cartopy.io import shapereader
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
from rasterio.crs import CRS
import requests

from climada.util.constants import CMAP_EXPOSURES, CMAP_CAT, CMAP_RASTER
from climada.util.files_handler import to_list
import climada.util.coordinates as u_coord

LOGGER = logging.getLogger(__name__)

RESOLUTION = 250
"""Number of pixels in one direction in rendered image"""

BUFFER = 1.0
"""Degrees to add in the border"""

MAX_BINS = 2000
"""Maximum number of bins in geo_bin_from_array"""


def geo_bin_from_array(array_sub, geo_coord, var_name, title,
                       pop_name=True, buffer=BUFFER, extend='neither',
                       proj=ccrs.PlateCarree(), shapes=True, axes=None,
                       figsize=(9, 13), adapt_fontsize=True, **kwargs):
    """Plot array values binned over input coordinates.

    Parameters
    ----------
    array_sub : np.array(1d or 2d) or list(np.array)
        Each array (in a row or in  the list) are values at each point in corresponding
        geo_coord that are binned in one subplot.
    geo_coord : 2d np.array or list(2d np.array)
        (lat, lon) for each point in a row. If one provided, the same grid is used for all
        subplots. Otherwise provide as many as subplots in array_sub.
    var_name : str or list(str)
        label to be shown in the colorbar. If one provided, the same is used for all subplots.
        Otherwise provide as many as subplots in array_sub.
    title : str or list(str)
        subplot title. If one provided, the same is used for all subplots.
        Otherwise provide as many as subplots in array_sub.
    pop_name : bool, optional
        add names of the populated places, by default True.
    buffer : float, optional
        border to add to coordinates, by default BUFFER
    extend : str, optional
        extend border colorbar with arrows.
        [ 'neither' | 'both' | 'min' | 'max' ], by default 'neither'
    proj : ccrs, optional
        coordinate reference system of the given data, by default ccrs.PlateCarree()
    shapes : bool, optional
        Overlay Earth's countries coastlines to matplotlib.pyplot axis.
        The default is True
    axes : Axes or ndarray(Axes), optional
        by default None
    figsize : tuple, optional
        figure size for plt.subplots, by default (9, 13)
    adapt_fontsize : bool, optional
        If set to true, the size of the fonts will be adapted to the size of the figure. Otherwise
        the default matplotlib font size is used. Default is True.

    **kwargs
        arbitrary keyword arguments for hexbin matplotlib function

    Returns
    -------
    cartopy.mpl.geoaxes.GeoAxesSubplot

    Raises
    ------
    ValueError:
        Input array size missmatch
    """
    return _plot_scattered_data("hexbin", array_sub, geo_coord, var_name, title,
                                pop_name=pop_name, buffer=buffer, extend=extend,
                                proj=proj, shapes=shapes, axes=axes,
                                figsize=figsize, adapt_fontsize=adapt_fontsize, **kwargs)


def geo_scatter_from_array(array_sub, geo_coord, var_name, title,
                           pop_name=False, buffer=BUFFER, extend='neither',
                           proj=ccrs.PlateCarree(), shapes=True, axes=None,
                           figsize=(9, 13), adapt_fontsize=True, **kwargs):
    """Plot array values at input coordinates.

    Parameters
    ----------
    array_sub : np.array(1d or 2d) or list(np.array)
        Each array (in a row or in  the list) are values at each point in corresponding
        geo_coord that are binned in one subplot.
    geo_coord : 2d np.array or list(2d np.array)
        (lat, lon) for each point in a row. If one provided, the same grid is used for all
        subplots. Otherwise provide as many as subplots in array_sub.
    var_name : str or list(str)
        label to be shown in the colorbar. If one provided, the same is used for all subplots.
        Otherwise provide as many as subplots in array_sub.
    title : str or list(str)
        subplot title. If one provided, the same is used for all subplots.
        Otherwise provide as many as subplots in array_sub.
    pop_name : bool, optional
        add names of the populated places, by default False.
    buffer : float, optional
        border to add to coordinates, by default BUFFER
    extend : str, optional
        extend border colorbar with arrows.
        [ 'neither' | 'both' | 'min' | 'max' ], by default 'neither'
    proj : ccrs, optional
        coordinate reference system of the given data, by default ccrs.PlateCarree()
    shapes : bool, optional
        Overlay Earth's countries coastlines to matplotlib.pyplot axis.
        The default is True
    axes : Axes or ndarray(Axes), optional
        by default None
    figsize : tuple, optional
        figure size for plt.subplots, by default (9, 13)
    adapt_fontsize : bool, optional
        If set to true, the size of the fonts will be adapted to the size of the figure. Otherwise
        the default matplotlib font size is used. Default is True.
    **kwargs
        arbitrary keyword arguments for scatter matplotlib function

    Returns
    -------
    cartopy.mpl.geoaxes.GeoAxesSubplot

    Raises
    ------
    ValueError:
        Input array size missmatch
    """
    return _plot_scattered_data("scatter", array_sub, geo_coord, var_name, title,
                                pop_name=pop_name, buffer=buffer, extend=extend,
                                proj=proj, shapes=shapes, axes=axes,
                                figsize=figsize, adapt_fontsize=adapt_fontsize, **kwargs)


def _plot_scattered_data(method, array_sub, geo_coord, var_name, title,
                         pop_name=False, buffer=BUFFER, extend='neither',
                         proj=ccrs.PlateCarree(), shapes=True, axes=None,
                         figsize=(9, 13), adapt_fontsize=True, **kwargs):
    """Function for internal use in `geo_scatter_from_array` (when called with method="scatter")
    and `geo_bin_from_array` (when called with method="hexbin"). See the docstrings of the
    respective functions for more information on the parameters."""

    # Generate array of values used in each subplot
    num_im, list_arr = _get_collection_arrays(array_sub)
    list_tit = to_list(num_im, title, 'title')
    list_name = to_list(num_im, var_name, 'var_name')
    list_coord = to_list(num_im, geo_coord, 'geo_coord')

    if 'cmap' not in kwargs:
        kwargs['cmap'] = CMAP_EXPOSURES

    if axes is None:
        proj_plot = proj
        if isinstance(proj, ccrs.PlateCarree):
            # use different projections for plot and data to shift the central lon in the plot
            xmin, xmax = u_coord.lon_bounds(np.concatenate([c[:, 1] for c in list_coord]))
            proj_plot = ccrs.PlateCarree(central_longitude=0.5 * (xmin + xmax))
        _, axes, fontsize = make_map(num_im, proj=proj_plot, figsize=figsize, adapt_fontsize=adapt_fontsize)
    else:
        fontsize = None
    axes_iter = axes
    if not isinstance(axes, np.ndarray):
        axes_iter = np.array([[axes]])

    # Generate each subplot
    for array_im, axis, tit, name, coord in \
    zip(list_arr, axes_iter.flatten(), list_tit, list_name, list_coord):
        if coord.shape[0] != array_im.size:
            raise ValueError("Size mismatch in input array: %s != %s." %
                             (coord.shape[0], array_im.size))

        # Binned image with coastlines
        if isinstance(proj, ccrs.PlateCarree):
            xmin, ymin, xmax, ymax = u_coord.latlon_bounds(coord[:, 0], coord[:, 1], buffer=buffer)
            extent = (xmin, xmax, ymin, ymax)
        else:
            extent = _get_borders(coord, buffer=buffer, proj_limits=proj.x_limits + proj.y_limits)
        axis.set_extent((extent), proj)

        if shapes:
            add_shapes(axis)
        if pop_name:
            add_populated_places(axis, extent, proj, fontsize)

        if method == "hexbin":
            if 'gridsize' not in kwargs:
                kwargs['gridsize'] = min(int(array_im.size / 2), MAX_BINS)
            mappable = axis.hexbin(coord[:, 1], coord[:, 0], C=array_im,
                                   transform=proj, **kwargs)
        else:
            mappable = axis.scatter(coord[:, 1], coord[:, 0], c=array_im,
                                    transform=proj, **kwargs)

        # Create colorbar in this axis
        cbax = make_axes_locatable(axis).append_axes(
            'right', size="6.5%", pad=0.1, axes_class=plt.Axes)
        cbar = plt.colorbar(mappable, cax=cbax, orientation='vertical', extend=extend)
        cbar.set_label(name)
        axis.set_title("\n".join(wrap(tit)))
        if fontsize:
            cbar.ax.tick_params(labelsize=fontsize)
            cbar.ax.yaxis.get_offset_text().set_fontsize(fontsize)
            for item in [axis.title, cbar.ax.xaxis.label, cbar.ax.yaxis.label]:
                item.set_fontsize(fontsize)
    plt.tight_layout()
    return axes


def geo_im_from_array(array_sub, coord, var_name, title,
                      proj=None, smooth=True, axes=None, figsize=(9, 13), adapt_fontsize=True,
                      **kwargs):
    """Image(s) plot defined in array(s) over input coordinates.

    Parameters
    ----------
    array_sub : np.array(1d or 2d) or list(np.array)
        Each array (in a row or in  the list) are values at each point in corresponding
        geo_coord that are ploted in one subplot.
    coord : 2d np.array
        (lat, lon) for each point in a row. The same grid is used for all subplots.
    var_name : str or list(str)
        label to be shown in the colorbar. If one provided, the same is used for all subplots.
        Otherwise provide as many as subplots in array_sub.
    title : str or list(str)
        subplot title. If one provided, the same is used for all subplots.
        Otherwise provide as many as subplots in array_sub.
    proj : ccrs, optional
        coordinate reference system used in coordinates, by default None
    smooth : bool, optional
        smooth plot to RESOLUTIONxRESOLUTION, by default True
    axes : Axes or ndarray(Axes), optional
        by default None
    figsize : tuple, optional
        figure size for plt.subplots, by default (9, 13)
    adapt_fontsize : bool, optional
        If set to true, the size of the fonts will be adapted to the size of the figure. Otherwise
        the default matplotlib font size is used. Default is True.
    **kwargs
        arbitrary keyword arguments for pcolormesh matplotlib function

    Returns
    -------
    cartopy.mpl.geoaxes.GeoAxesSubplot

    Raises
    ------
    ValueError
    """

    # Generate array of values used in each subplot
    num_im, list_arr = _get_collection_arrays(array_sub)
    list_tit = to_list(num_im, title, 'title')
    list_name = to_list(num_im, var_name, 'var_name')
    list_coord = to_list(num_im, coord, 'geo_coord')


    is_reg, height, width = u_coord.grid_is_regular(coord)
    extent = _get_borders(coord, proj_limits=(-360, 360, -90, 90))
    mid_lon = 0
    if not proj:
        mid_lon = 0.5 * sum(extent[:2])
        proj = ccrs.PlateCarree(central_longitude=mid_lon)
    if 'vmin' not in kwargs:
        kwargs['vmin'] = np.nanmin(array_sub)
    if 'vmax' not in kwargs:
        kwargs['vmax'] = np.nanmax(array_sub)
    if axes is None:
        if isinstance(proj, ccrs.PlateCarree):
            # use different projections for plot and data to shift the central lon in the plot
            xmin, xmax = u_coord.lon_bounds(np.concatenate([c[:, 1] for c in list_coord]))
            proj_plot = ccrs.PlateCarree(central_longitude=0.5 * (xmin + xmax))
        _, axes, fontsize = make_map(num_im, proj=proj_plot, figsize=figsize, adapt_fontsize=adapt_fontsize)
    else:
        fontsize = None
    axes_iter = axes
    if not isinstance(axes, np.ndarray):
        axes_iter = np.array([[axes]])

    if 'cmap' not in kwargs:
        kwargs['cmap'] = CMAP_RASTER

    # Generate each subplot
    for array_im, axis, tit, name in zip(list_arr, axes_iter.flatten(), list_tit, list_name):
        if coord.shape[0] != array_im.size:
            raise ValueError("Size mismatch in input array: %s != %s." %
                             (coord.shape[0], array_im.size))
        if smooth or not is_reg:
            # Create regular grid where to interpolate the array
            grid_x, grid_y = np.mgrid[
                extent[0]: extent[1]: complex(0, RESOLUTION),
                extent[2]: extent[3]: complex(0, RESOLUTION)]
            grid_im = griddata((coord[:, 1], coord[:, 0]), array_im,
                               (grid_x, grid_y))
        else:
            grid_x = coord[:, 1].reshape((width, height)).transpose()
            grid_y = coord[:, 0].reshape((width, height)).transpose()
            grid_im = np.array(array_im.reshape((width, height)).transpose())
            if grid_y[0, 0] > grid_y[0, -1]:
                grid_y = np.flip(grid_y)
                grid_im = np.flip(grid_im, 1)
            grid_im = np.resize(grid_im, (height, width, 1))
        axis.set_extent((extent[0] - mid_lon, extent[1] - mid_lon,
                         extent[2], extent[3]), crs=proj)

        # Add coastline to axis
        add_shapes(axis)
        # Create colormesh, colorbar and labels in axis
        cbax = make_axes_locatable(axis).append_axes('right', size="6.5%",
                                                     pad=0.1, axes_class=plt.Axes)
        img = axis.pcolormesh(grid_x - mid_lon, grid_y, np.squeeze(grid_im),
                              transform=proj, **kwargs)
        cbar = plt.colorbar(img, cax=cbax, orientation='vertical')
        cbar.set_label(name)
        axis.set_title("\n".join(wrap(tit)))
        if fontsize:
            cbar.ax.tick_params(labelsize=fontsize)
            cbar.ax.yaxis.get_offset_text().set_fontsize(fontsize)
            for item in [axis.title, cbar.ax.xaxis.label, cbar.ax.yaxis.label]:
                item.set_fontsize(fontsize)

    plt.tight_layout()
    return axes


def geo_scatter_categorical(array_sub, geo_coord, var_name, title,
                            cat_name=None, adapt_fontsize=True, **kwargs):
    """
    Map plots for categorical data defined in array(s) over input
    coordinates. The categories must be a finite set of unique values
    as can be identified by np.unique() (mix of int, float, strings, ...).

    The categories are shared among all subplots, i.e. are obtained from
    np.unique(array_sub).
    Eg.:
        array_sub = [[1, 2, 1.0, 2], [1, 2, 'a', 'a']]
        -> categories mapping is [[0, 2, 1, 2], [0, 2, 3, 3]]

    Same category: 1 and '1'
    Different categories: 1 and 1.0

    This method wraps around util.geo_scatter_from_array and uses
    all its args and kwargs.

    Parameters
    ----------
    array_sub : np.array(1d or 2d) or list(np.array)
        Each array (in a row or in  the list) are values at each point
        in corresponding geo_coord that are binned in one subplot.
    geo_coord : 2d np.array or list(2d np.array)
        (lat, lon) for each point in a row. If one provided, the same grid
        is used for all subplots. Otherwise provide as many as subplots
        in array_sub.
    var_name : str or list(str)
        label to be shown in the colorbar. If one
        provided, the same is used for all subplots. Otherwise provide as
        many as subplots in array_sub.
    title : str or list(str)
        subplot title. If one provided, the same is
        used for all subplots. Otherwise provide as many as subplots in
        array_sub.
    cat_name : dict, optional
        Categories name for the colorbar labels.
        Keys are all the unique values in array_sub, values are their labels.
        The default is labels = unique values.
    adapt_fontsize : bool, optional
        If set to true, the size of the fonts will be adapted to the size of the figure. Otherwise
        the default matplotlib font size is used. Default is True.
    **kwargs
        Arbitrary keyword arguments for hexbin matplotlib function

    Returns
    -------
    cartopy.mpl.geoaxes.GeoAxesSubplot

    """

    # convert sorted categories to numeric array [0, 1, ...]
    array_sub = np.array(array_sub)
    array_sub_unique, array_sub_cat = np.unique(array_sub, return_inverse=True) #flattens array
    array_sub_cat = array_sub_cat.reshape(array_sub.shape)
    array_sub_n = array_sub_unique.size

    if 'cmap' in kwargs:
        # optional user defined colormap (can be continuous)
        cmap = kwargs['cmap']
        if isinstance(cmap, str):
            cmap_name = cmap
            cmap = mpl.cm.get_cmap(cmap)
        else:
            cmap_name = 'defined by the user'
    else:
        # default qualitative colormap
        cmap_name = CMAP_CAT
        cmap = mpl.colors.ListedColormap(
            plt.get_cmap(cmap_name).colors[:array_sub_n]
            )

    if array_sub_n > cmap.N:
        LOGGER.warning("More than %d categories cannot be plotted accurately "
                       "using the colormap %s. Please specify "
                       "a different qualitative colormap using the `cmap` "
                       "attribute. For Matplotlib's built-in colormaps, see "
           "https://matplotlib.org/stable/tutorials/colors/colormaps.html",
                       cmap.N, cmap_name)

    # define the discrete colormap kwargs
    kwargs['cmap'] = mpl.cm.get_cmap(cmap, array_sub_n)
    kwargs['vmin'] = -0.5
    kwargs['vmax'] = array_sub_n - 0.5

    # #create the axes
    axes = _plot_scattered_data(
        "scatter", array_sub_cat, geo_coord, var_name, title,
        adapt_fontsize=adapt_fontsize, **kwargs)

    #add colorbar labels
    if cat_name is None:
        cat_name = array_sub_unique.astype(str)
    if not isinstance(cat_name, dict):
        cat_name = dict(zip(array_sub_unique, cat_name))
    cat_name = {str(key): value for key, value in cat_name.items()}
    if not isinstance(axes, np.ndarray):
        axes = np.array([axes])
    for ax in axes.ravel():
        cbar = [coll.colorbar for coll in ax.collections if coll.colorbar is not None]
        if len(cbar) > 0:
            cbar = cbar[-1]
            cbar.set_ticks(np.arange(array_sub_n))
            cbar.set_ticklabels([cat_name[str(val)] for val in array_sub_unique])

    return axes


def make_map(num_sub=1, figsize=(9, 13), proj=ccrs.PlateCarree(), adapt_fontsize=True):
    """
    Create map figure with cartopy.

    Parameters
    ----------
    num_sub : int or tuple
        number of total subplots in figure OR number of
        subfigures in row and column: (num_row, num_col).
    figsize : tuple
        figure size for plt.subplots
    proj : cartopy.crs projection, optional
        geographical projection,
        The default is PlateCarree default.
    adapt_fontsize : bool, optional
        If set to true, the size of the fonts will be adapted to the size of the figure. Otherwise
        the default matplotlib font size is used. Default is True.

    Returns
    -------
    fig, axis_sub : matplotlib.figure.Figure, cartopy.mpl.geoaxes.GeoAxesSubplot
    """
    if isinstance(num_sub, int):
        num_row, num_col = _get_row_col_size(num_sub)
    else:
        num_row, num_col = num_sub

    fig, axis_sub = plt.subplots(num_row, num_col, figsize=figsize,
                                 subplot_kw=dict(projection=proj))
    axes_iter = axis_sub
    if not isinstance(axis_sub, np.ndarray):
        axes_iter = np.array([[axis_sub]])

    for axis in axes_iter.flatten():
        try:
            grid = axis.gridlines(draw_labels=True, alpha=0.2, transform=proj)
            grid.top_labels = grid.right_labels = False
            grid.xformatter = LONGITUDE_FORMATTER
            grid.yformatter = LATITUDE_FORMATTER
            if adapt_fontsize:
                fontsize = axis.bbox.width/35
                if fontsize < 10:
                    fontsize = 10
                grid.xlabel_style = {'size': fontsize}
                grid.ylabel_style = {'size': fontsize}
            else:
                fontsize = None
        except TypeError:
            pass

    if num_col > 1:
        fig.subplots_adjust(wspace=0.3)
    if num_col > 2:
        fig.subplots_adjust(wspace=0.5)
    if num_row > 1:
        fig.subplots_adjust(hspace=-0.5)

    return fig, axis_sub, fontsize

def add_shapes(axis):
    """
    Overlay Earth's countries coastlines to matplotlib.pyplot axis.

    Parameters
    ----------
    axis : cartopy.mpl.geoaxes.GeoAxesSubplot
        Cartopy axis
    projection : cartopy.crs projection, optional
        Geographical projection.
        The default is PlateCarree.
    """
    shp_file = shapereader.natural_earth(resolution='10m', category='cultural',
                                         name='admin_0_countries')
    shp = shapereader.Reader(shp_file)
    for geometry in shp.geometries():
        axis.add_geometries([geometry], crs=ccrs.PlateCarree(), facecolor='none',
                            edgecolor='dimgray')

def _ensure_utf8(val):
    # Without the `*.cpg` file present, the shape reader wrongly assumes latin-1 encoding:
    # https://github.com/SciTools/cartopy/issues/1282
    # https://github.com/SciTools/cartopy/commit/6d787b01e122eea68b67a9b2966e45877755a52d
    # As a workaround, we encode and decode again, unless this fails which means
    # that the `*.cpg` is present and the encoding is correct:
    try:
        return val.encode('latin-1').decode('utf-8')
    except:
        return val

def add_populated_places(axis, extent, proj=ccrs.PlateCarree(), fontsize=None):
    """
    Add city names.

    Parameters
    ----------
    axis : cartopy.mpl.geoaxes.GeoAxesSubplot
        cartopy axis.
    extent : list
        geographical limits [min_lon, max_lon, min_lat, max_lat]
    proj : cartopy.crs projection, optional
        geographical projection,
        The default is PlateCarree.
    fontsize : int, optional
        Size of the fonts. If set to None, the default matplotlib settings
        are used.

    """
    shp_file = shapereader.natural_earth(resolution='50m', category='cultural',
                                         name='populated_places_simple')

    shp = shapereader.Reader(shp_file)
    ext_pts = list(box(extent[0], extent[2], extent[1], extent[3]).exterior.coords)
    ext_trans = [ccrs.PlateCarree().transform_point(pts[0], pts[1], proj)
                 for pts in ext_pts]
    for rec, point in zip(shp.records(), shp.geometries()):
        if ext_trans[2][0] < point.x <= ext_trans[0][0]:
            if ext_trans[0][1] < point.y <= ext_trans[1][1]:
                axis.plot(point.x, point.y, color='navy', marker='o',
                          transform=ccrs.PlateCarree(), markerfacecolor='None')
                axis.text(point.x, point.y, _ensure_utf8(rec.attributes['name']),
                          horizontalalignment='right', verticalalignment='bottom',
                          transform=ccrs.PlateCarree(), color='navy', fontsize=fontsize)

def add_cntry_names(axis, extent, proj=ccrs.PlateCarree(), fontsize=None):
    """
    Add country names.

    Parameters
    ----------
    axis : cartopy.mpl.geoaxes.GeoAxesSubplot
        Cartopy axis.
    extent : list
        geographical limits [min_lon, max_lon, min_lat, max_lat]
    proj : cartopy.crs projection, optional
        Geographical projection.
        The default is PlateCarree.
     fontsize : int, optional
        Size of the fonts. If set to None, the default matplotlib settings
        are used.
    """
    shp_file = shapereader.natural_earth(resolution='10m', category='cultural',
                                         name='admin_0_countries')

    shp = shapereader.Reader(shp_file)
    ext_pts = list(box(extent[0], extent[2], extent[1], extent[3]).exterior.coords)
    ext_trans = [ccrs.PlateCarree().transform_point(pts[0], pts[1], proj)
                 for pts in ext_pts]
    for rec, point in zip(shp.records(), shp.geometries()):
        point_x = point.centroid.xy[0][0]
        point_y = point.centroid.xy[1][0]
        if ext_trans[2][0] < point_x <= ext_trans[0][0]:
            if ext_trans[0][1] < point_y <= ext_trans[1][1]:
                axis.text(point_x, point_y, _ensure_utf8(rec.attributes['NAME']),
                          horizontalalignment='center', verticalalignment='center',
                          transform=ccrs.PlateCarree(), fontsize=fontsize, color='navy')

def _get_collection_arrays(array_sub):
    """
    Get number of array rows and generate list of array if only one row

    Parameters
    ----------
    array_sub : np.array(1d or 2d) or list(np.array)
        Each array (in a row
        or in  the list) are values at each point in corresponding

    Returns
    -------
    num_im, list_arr : int, 2d np.ndarray or list(1d np.array)
        Number of array rows and list of array
    """
    num_im = 1
    if not isinstance(array_sub, list):
        if len(array_sub.shape) == 1 or array_sub.shape[1] == 1:
            list_arr = list()
            list_arr.append(array_sub)
        else:
            list_arr = array_sub
            num_im = array_sub.shape[0]
    else:
        num_im = len(array_sub)
        list_arr = array_sub

    return num_im, list_arr

def _get_row_col_size(num_sub):
    """
    Compute number of rows and columns of subplots in figure.

    Parameters
    ----------
    num_sub : int
        number of subplots

    Returns
    -------
    num_row, num_col : int, int
        Number of rows and columns
    """
    if num_sub <= 3:
        num_col = num_sub
        num_row = 1
    else:
        if num_sub % 3 == 0:
            num_col = 3
            num_row = int(num_sub / 3)
        else:
            num_col = 2
            num_row = int(num_sub / 2) + num_sub % 2
    return num_row, num_col

def _get_borders(geo_coord, buffer=0, proj_limits=(-180, 180, -90, 90)):
    """
    Get min and max longitude and min and max latitude (in this order).

    Parameters
    ----------
    geo_coord : 2d np.array
        (lat, lon) for each point in a row.
    buffer : float, optional
        border to add. The default is 0
    proj_limits  : tuple, optional
        limits of geographical projection (lon_min, lon_max, lat_min, lat_max)
        The default is (-180, 180, -90, 90)

    Returns
    -------
    extent : list [min_lon, max_lon, min_lat, max_lat]
    """
    min_lon = max(np.min(geo_coord[:, 1]) - buffer, proj_limits[0])
    max_lon = min(np.max(geo_coord[:, 1]) + buffer, proj_limits[1])
    min_lat = max(np.min(geo_coord[:, 0]) - buffer, proj_limits[2])
    max_lat = min(np.max(geo_coord[:, 0]) + buffer, proj_limits[3])
    return [min_lon, max_lon, min_lat, max_lat]

def get_transformation(crs_in):
    """
    Get projection and its units to use in cartopy transforamtions from
    current crs

    Parameters
    ----------
    crs_in : str
        Current crs

    Returns
    ------
    crs_epsg : ccrs.Projection
    units : str
    """
    try:
        if CRS.from_user_input(crs_in) == CRS.from_user_input('EPSG:3395'):
            crs_epsg = ccrs.Mercator()
        else:
            crs_epsg = ccrs.epsg(CRS.from_user_input(crs_in).to_epsg())
    except ValueError:
        crs_epsg = ccrs.PlateCarree()
    except requests.exceptions.ConnectionError:
        LOGGER.warning('No internet connection.'
                       ' Using projection PlateCarree in plot.')
        crs_epsg = ccrs.PlateCarree()

    try:
        units = (crs_epsg.proj4_params.get('units')
                # As of cartopy 0.20 the proj4_params attribute is {} for CRS from an EPSG number
                # (see issue raised https://github.com/SciTools/cartopy/issues/1974
                # and longterm discussion on https://github.com/SciTools/cartopy/issues/813).
                # In these cases the units can be fetched through the method `to_dict`.
                or crs_epsg.to_dict().get('units', '°'))
    except AttributeError:
                # This happens in setups with cartopy<0.20, where `to_dict` is not defined.
                # Officially, we require cartopy>=0.20, but there are still users around that
                # can't upgrade due to https://github.com/SciTools/iris/issues/4468
        units = '°'
    return crs_epsg, units


def multibar_plot(ax, data, colors=None, total_width=0.8, single_width=1,
                  legend=True, ticklabels=None, invert_axis=False):
    """
    Draws a bar plot with multiple bars per data point.
    https://stackoverflow.com/questions/14270391/python-matplotlib-multiple-bars

    Parameters
    ----------
    ax : matplotlib.pyplot.axis
        The axis we want to draw our plot on.

    data: dictionary
        A dictionary containing the data we want to plot. Keys are the names of the
        data, the items is a list of the values.

        Example:
        data = {
            "x": [1, 2, 3],
            "y": [1, 2, 3],
            "z": [1, 2, 3],
        }
        fig, ax = plt.subplots()
        multibar_plot(ax, data, xticklabels=["a", "b", "c"])


    colors : array-like, optional
        A list of colors which are used for the bars. If None, the colors
        will be the standard matplotlib color cyle. (default: None)

    total_width : float, optional, default: 0.8
        The width of a bar group. 0.8 means that 80% of the x-axis is covered
        by bars and 20% will be spaces between the bars.

    single_width: float, optional, default: 1
        The relative width of a single bar within a group. 1 means the bars
        will touch eachother within a group, values less than 1 will make
        these bars thinner.

    legend: bool, optional, default: True
        If this is set to true, a legend will be added to the axis.

    ticklabels: list, optional, default: None
        labels of the xticks (yticks if invert_axis=True)

    invert_axis: boolean, default: False
        Invert the x and y axis. By default, the bars are vertical.
        invert_axis=True gives horizontal bars.
    """

    # Check if colors where provided, otherwhise use the default color cycle
    if colors is None:
        colors = plt.rcParams['axes.prop_cycle'].by_key()['color']

    # Number of bars per group
    n_bars = len(data)

    # The width of a single bar
    bar_width = total_width / n_bars

    # List containing handles for the drawn bars, used for the legend
    bars = []

    # Iterate over all data
    for i, (_name, values) in enumerate(data.items()):
        # The offset in x direction of that bar
        x_offset = (i - n_bars / 2) * bar_width + bar_width / 2

        # Draw a bar for every value of that type
        for x, y in enumerate(values):
            if invert_axis:
                bar = ax.barh(x + x_offset, width=y, height=bar_width * single_width,
                              color=colors[i % len(colors)])
            else:
                bar = ax.bar(x + x_offset, y, width=bar_width * single_width,
                             color=colors[i % len(colors)])

        # Add a handle to the last drawn bar, which we'll need for the legend
        bars.append(bar[0])

    if ticklabels:
        if invert_axis:
            plt.setp(ax, yticks=range(len(data)), yticklabels=ticklabels)
        else:
            plt.setp(ax, xticks=range(len(data)), xticklabels=ticklabels)

    # Draw legend if we need
    if legend:
        ax.legend(bars, data.keys())



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define ImpactFunc class.
"""

__all__ = ['ImpactFunc']

import logging
import numpy as np
import matplotlib.pyplot as plt

import climada.util.checker as u_check

LOGGER = logging.getLogger(__name__)

class ImpactFunc():
    """Contains the definition of one impact function.

    Attributes
    ----------
    haz_type : str
        hazard type acronym (e.g. 'TC')
    id : int or str
        id of the impact function. Exposures of the same type
        will refer to the same impact function id
    name : str
        name of the ImpactFunc
    intensity_unit : str
        unit of the intensity
    intensity : np.array
        intensity values
    mdd : np.array
        mean damage (impact) degree for each intensity (numbers
        in [0,1])
    paa : np.array
        percentage of affected assets (exposures) for each
        intensity (numbers in [0,1])
    """
    def __init__(self):
        """Empty initialization."""
        self.id = ''
        self.name = ''
        self.intensity_unit = ''
        self.haz_type = ''
        # Followng values defined for each intensity value
        self.intensity = np.array([])
        self.mdd = np.array([])
        self.paa = np.array([])

    def calc_mdr(self, inten):
        """Interpolate impact function to a given intensity.

        Parameters
        ----------
        inten : float or np.array
            intensity, the x-coordinate of the
            interpolated values.

        Returns
        -------
        np.array
        """
#        return np.interp(inten, self.intensity, self.mdd * self.paa)
        return np.interp(inten, self.intensity, self.paa) * \
            np.interp(inten, self.intensity, self.mdd)

    def plot(self, axis=None, **kwargs):
        """Plot the impact functions MDD, MDR and PAA in one graph, where
        MDR = PAA * MDD.

        Parameters
        ----------
        axis : matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        kwargs : optional
            arguments for plot matplotlib function, e.g. marker='x'

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
        """
        if axis is None:
            _, axis = plt.subplots(1, 1)

        title = '%s %s' % (self.haz_type, str(self.id))
        if self.name != str(self.id):
            title += ': %s' % self.name
        axis.set_xlabel('Intensity (' + self.intensity_unit + ')')
        axis.set_ylabel('Impact (%)')
        axis.set_title(title)
        axis.plot(self.intensity, self.mdd * 100, 'b', label='MDD', **kwargs)
        axis.plot(self.intensity, self.paa * 100, 'r', label='PAA', **kwargs)
        axis.plot(self.intensity, self.mdd * self.paa * 100, 'k--', label='MDR', **kwargs)

        axis.set_xlim((self.intensity.min(), self.intensity.max()))
        axis.legend()
        return axis

    def check(self):
        """Check consistent instance data.

        Raises
        ------
        ValueError
        """
        num_exp = len(self.intensity)
        u_check.size(num_exp, self.mdd, 'ImpactFunc.mdd')
        u_check.size(num_exp, self.paa, 'ImpactFunc.paa')

        if num_exp == 0:
            LOGGER.warning("%s impact function with name '%s' (id=%s) has empty"
                           " intensity.", self.haz_type, self.name, self.id)
            return

        # Warning for non-vanishing impact at intensity 0. If positive
        # and negative intensity warning for interpolation at intensity 0.
        zero_idx = np.where(self.intensity == 0)[0]
        if zero_idx.size != 0:
            if self.mdd[zero_idx[0]] != 0 or self.paa[zero_idx[0]] != 0:
                LOGGER.warning('For intensity = 0, mdd != 0 or paa != 0. '
                               'Consider shifting the origin of the intensity '
                               'scale. In impact.calc the impact is always '
                               'null at intensity = 0.')
        elif self.intensity[0] < 0 and self.intensity[-1] > 0:
            LOGGER.warning('Impact function might be interpolated to non-zero'
                           ' value at intensity = 0. Consider shifting the '
                           'origin of the intensity scale. In impact.calc '
                           'the impact is always null at intensity = 0.')

    @classmethod
    def from_step_impf(cls, intensity, mdd=(0, 1), paa=(1, 1), impf_id=1):

        """ Step function type impact function.

        By default, everything is destroyed above the step.
        Useful for high resolution modelling.

        This method modifies self (climada.entity.impact_funcs instance)
        by assigning an id, intensity, mdd and paa to the impact function.

        Parameters
        ----------
        intensity: tuple(float, float, float)
            tuple of 3-intensity numbers: (minimum, threshold, maximum)
        mdd: tuple(float, float)
            (min, max) mdd values. The default is (0, 1)
        paa: tuple(float, float)
            (min, max) paa values. The default is (1, 1)
        impf_id : int, optional, default=1
            impact function id

        Return
        ------
        impf : climada.entity.impact_funcs.ImpactFunc
            Step impact function

        """

        impf = cls()
        impf.id = impf_id
        inten_min, threshold, inten_max = intensity
        impf.intensity = np.array([inten_min, threshold, threshold, inten_max])
        paa_min, paa_max = paa
        impf.paa = np.array([paa_min, paa_min, paa_max, paa_max])
        mdd_min, mdd_max = mdd
        impf.mdd = np.array([mdd_min, mdd_min, mdd_max, mdd_max])

        return impf

    def set_step_impf(self, *args, **kwargs):
        """This function is deprecated, use ImpactFunc.from_step_impf instead."""
        LOGGER.warning("The use of ImpactFunc.set_step_impf is deprecated." +
                        "Use ImpactFunc.from_step_impf instead.")
        self.__dict__ = ImpactFunc.from_step_impf(*args, **kwargs).__dict__

    @classmethod
    def from_sigmoid_impf(cls, intensity, L, k, x0, if_id=1):
        """ Sigmoid type impact function hinging on three parameter.

        This type of impact function is very flexible for any sort of study,
        hazard and resolution. The sigmoid is defined as:

        .. math::
            f(x) = \frac{L}{1+exp^{-k(x-x0)}}

        For more information: https://en.wikipedia.org/wiki/Logistic_function

        This method modifies self (climada.entity.impact_funcs instance)
        by assining an id, intensity, mdd and paa to the impact function.

        Parameters
        ----------
            intensity: tuple(float, float, float)
                tuple of 3 intensity numbers along np.arange(min, max, step)
            L : float
                "top" of sigmoid
            k : float
                "slope" of sigmoid
            x0 : float
                intensity value where f(x)==L/2
            if_id : int, optional, default=1
                impact function id

        Return
        ------
        impf : climada.entity.impact_funcs.ImpactFunc
            Step impact function

        """
        impf = cls()
        impf.id = if_id
        inten_min, inten_max, inten_step = intensity
        impf.intensity = np.arange(inten_min, inten_max, inten_step)
        impf.paa = np.ones(len(impf.intensity))
        impf.mdd = L / (1 + np.exp(-k * (impf.intensity - x0)))

        return impf

    def set_sigmoid_impf(self, *args, **kwargs):
        """This function is deprecated, use LitPop.from_countries instead."""
        LOGGER.warning("The use of ImpactFunc.set_sigmoid_impf is deprecated."
                       "Use ImpactFunc.from_sigmoid_impf instead.")
        self.__dict__ = ImpactFunc.from_sigmoid_impf(*args, **kwargs).__dict__





"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test Exposures from MATLAB file.
"""
import unittest
import copy

from climada import CONFIG
from climada.entity.exposures.base import Exposures, DEF_VAR_MAT

ENT_TEST_MAT = CURR_DIR = CONFIG.exposures.test_data.dir().joinpath('demo_today.mat')

class TestReader(unittest.TestCase):
    """Test reader functionality of the ExposuresMat class"""

    def test_read_demo_pass(self):
        """Read one single excel file"""
        # Read demo excel file
        expo = Exposures.from_mat(ENT_TEST_MAT)

        # Check results
        n_expos = 50

        self.assertEqual(expo.gdf.index.shape, (n_expos,))
        self.assertEqual(expo.gdf.index[0], 0)
        self.assertEqual(expo.gdf.index[n_expos - 1], n_expos - 1)

        self.assertEqual(expo.gdf.value.shape, (n_expos,))
        self.assertEqual(expo.gdf.value[0], 13927504367.680632)
        self.assertEqual(expo.gdf.value[n_expos - 1], 12624818493.687229)

        self.assertEqual(expo.gdf.deductible.shape, (n_expos,))
        self.assertEqual(expo.gdf.deductible[0], 0)
        self.assertEqual(expo.gdf.deductible[n_expos - 1], 0)

        self.assertEqual(expo.gdf.cover.shape, (n_expos,))
        self.assertEqual(expo.gdf.cover[0], 13927504367.680632)
        self.assertEqual(expo.gdf.cover[n_expos - 1], 12624818493.687229)

        self.assertIn('int', str(expo.gdf.impf_.dtype))
        self.assertEqual(expo.gdf.impf_.shape, (n_expos,))
        self.assertEqual(expo.gdf.impf_[0], 1)
        self.assertEqual(expo.gdf.impf_[n_expos - 1], 1)

        self.assertIn('int', str(expo.gdf.category_id.dtype))
        self.assertEqual(expo.gdf.category_id.shape, (n_expos,))
        self.assertEqual(expo.gdf.category_id[0], 1)
        self.assertEqual(expo.gdf.category_id[n_expos - 1], 1)

        self.assertIn('int', str(expo.gdf.centr_.dtype))
        self.assertEqual(expo.gdf.centr_.shape, (n_expos,))
        self.assertEqual(expo.gdf.centr_[0], 47)
        self.assertEqual(expo.gdf.centr_[n_expos - 1], 46)

        self.assertTrue('region_id' not in expo.gdf)

        self.assertEqual(expo.gdf.latitude.shape, (n_expos,))
        self.assertEqual(expo.gdf.latitude[0], 26.93389900000)
        self.assertEqual(expo.gdf.latitude[n_expos - 1], 26.34795700000)
        self.assertEqual(expo.gdf.longitude[0], -80.12879900000)
        self.assertEqual(expo.gdf.longitude[n_expos - 1], -80.15885500000)

        self.assertEqual(expo.ref_year, 2016)
        self.assertEqual(expo.value_unit, 'USD')
        self.assertEqual(expo.tag.file_name, str(ENT_TEST_MAT))

class TestObligatories(unittest.TestCase):
    """Test reading exposures obligatory values."""

    def test_no_value_fail(self):
        """Error if no values."""
        new_var_names = copy.deepcopy(DEF_VAR_MAT)
        new_var_names['var_name']['val'] = 'no valid value'
        with self.assertRaises(KeyError):
            Exposures.from_mat(ENT_TEST_MAT, var_names=new_var_names)

    def test_no_impact_fail(self):
        """Error if no impact ids."""
        new_var_names = copy.deepcopy(DEF_VAR_MAT)
        new_var_names['var_name']['impf'] = 'no valid value'
        with self.assertRaises(KeyError):
            Exposures.from_mat(ENT_TEST_MAT, var_names=new_var_names)

    def test_no_coord_fail(self):
        """Error if no coordinates."""
        new_var_names = copy.deepcopy(DEF_VAR_MAT)
        new_var_names['var_name']['lat'] = 'no valid Latitude'
        with self.assertRaises(KeyError):
            Exposures.from_mat(ENT_TEST_MAT, var_names=new_var_names)

        new_var_names['var_name']['lat'] = 'nLatitude'
        new_var_names['var_name']['lon'] = 'no valid Longitude'
        with self.assertRaises(KeyError):
            Exposures.from_mat(ENT_TEST_MAT, var_names=new_var_names)

class TestOptionals(unittest.TestCase):
    """Test reading exposures optional values."""

    def test_no_category_pass(self):
        """Not error if no category id."""
        new_var_names = copy.deepcopy(DEF_VAR_MAT)
        new_var_names['var_name']['cat'] = 'no valid category'
        exp = Exposures.from_mat(ENT_TEST_MAT, var_names=new_var_names)

        # Check results
        self.assertTrue('category_id' not in exp.gdf)

    def test_no_region_pass(self):
        """Not error if no region id."""
        new_var_names = copy.deepcopy(DEF_VAR_MAT)
        new_var_names['var_name']['reg'] = 'no valid region'
        exp = Exposures.from_mat(ENT_TEST_MAT, var_names=new_var_names)

        # Check results
        self.assertTrue('region_id' not in exp.gdf)

    def test_no_unit_pass(self):
        """Not error if no value unit."""
        new_var_names = copy.deepcopy(DEF_VAR_MAT)
        new_var_names['var_name']['uni'] = 'no valid unit'
        exp = Exposures.from_mat(ENT_TEST_MAT, var_names=new_var_names)

        # Check results
        self.assertEqual('USD', exp.value_unit)

    def test_no_assigned_pass(self):
        """Not error if no value unit."""
        new_var_names = copy.deepcopy(DEF_VAR_MAT)
        new_var_names['var_name']['ass'] = 'no valid assign'
        exp = Exposures.from_mat(ENT_TEST_MAT, var_names=new_var_names)

        # Check results
        self.assertTrue('centr_' not in exp.gdf)

    def test_no_refyear_pass(self):
        """Not error if no value unit."""
        new_var_names = copy.deepcopy(DEF_VAR_MAT)
        new_var_names['var_name']['ref'] = 'no valid ref'
        exp = Exposures.from_mat(ENT_TEST_MAT, var_names=new_var_names)

        # Check results
        self.assertEqual(2018, exp.ref_year)

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestReader)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestOptionals))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestObligatories))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test of finance module
"""
import unittest
import numpy as np
from cartopy.io import shapereader

from climada.util.finance import net_present_value, gdp, income_group, \
nat_earth_adm0, world_bank, wealth2gdp, world_bank_wealth_account, _gdp_twn

SHP_FN = shapereader.natural_earth(resolution='10m', category='cultural',
                                   name='admin_0_countries')
SHP_FILE = shapereader.Reader(SHP_FN)

class TestNetpresValue(unittest.TestCase):
    """Test date functions"""
    def test_net_pres_val_pass(self):
        """Test net_present_value against MATLAB reference"""
        years = np.arange(2018, 2041)
        disc_rates = np.ones(years.size) * 0.02
        val_years = np.ones(years.size) * 6.512201157564418e9
        res = net_present_value(years, disc_rates, val_years)

        self.assertEqual(1.215049630691397e+11, res)

class TestWBData(unittest.TestCase):
    """Test World Bank data"""
    def test_ne_income_grp_aia_pass(self):
        """Test nat_earth_adm0 function Anguilla."""
        ref_year = 2012
        res_year, res_val = nat_earth_adm0('AIA', 'INCOME_GRP',
                                           shp_file=SHP_FILE)

        ref_year = 0
        ref_val = 3
        self.assertEqual(res_year, ref_year)
        self.assertEqual(res_val, ref_val)

    def test_wb_income_grp_sxm_pass(self):
        """Test world_bank function Sint Maarten."""
        ref_year = 2012
        res_year, res_val = world_bank('SXM', ref_year, 'INC_GRP')

        ref_year = 2012
        ref_val = 4
        self.assertEqual(res_year, ref_year)
        self.assertEqual(res_val, ref_val)

    def test_income_grp_sxm_1999_pass(self):
        """Test income_group function Sint Maarten."""
        ref_year = 1999
        with self.assertLogs('climada.util.finance', level='INFO') as cm:
            res_year, res_val = income_group('SXM', ref_year, SHP_FILE)

        ref_year = 2010
        ref_val = 4
        self.assertIn('Income group SXM 2010: 4.', cm.output[0])
        self.assertEqual(res_year, ref_year)
        self.assertEqual(res_val, ref_val)

    def test_ne_gdp_aia_2012_pass(self):
        """Test nat_earth_adm0 function Anguilla."""
        ref_year = 2012
        res_year, res_val = nat_earth_adm0('AIA', 'GDP_MD',
                                           'GDP_YEAR', SHP_FILE)

        ref_year = 2009
        ref_val = 1.75e+08
        self.assertEqual(res_year, ref_year)
        self.assertEqual(res_val, ref_val)

    def test_gdp_sxm_2010_pass(self):
        """Test gdp function Sint Maarten."""
        # If World Bank input data changes, make sure to set ref_year to a year where
        # no data is available so that the next available data point has to be selected.
        ref_year = 2010
        with self.assertLogs('climada.util.finance', level='INFO') as cm:
            res_year, res_val = gdp('SXM', ref_year)

        ref_val = 936089385.47486 # reference GDP value
        ref_year = 2011 # nearest year with data available (might change)
        # GDP and years with data available might change if worldbank input
        # data changes, check magnitude and adjust ref_val and/or ref_year
        # if test fails:
        self.assertIn('GDP SXM %i: %1.3e' % (ref_year, ref_val), cm.output[0])
        self.assertEqual(res_year, ref_year)
        self.assertAlmostEqual(res_val, ref_val, places=0)

    def test_gdp_twn_2012_pass(self):
        """Test gdp function TWN."""
        ref_year = 2014
        res_year, res_val = gdp('TWN', ref_year)
        _, res_val_direct = _gdp_twn(ref_year)
        ref_val = 530515000000.0
        ref_year = 2014
        self.assertEqual(res_year, ref_year)
        self.assertEqual(res_val, ref_val)
        self.assertEqual(res_val_direct, ref_val)


    def test_wb_esp_1950_pass(self):
        """Test world_bank function Sint Maarten."""
        ref_year = 1950
        res_year, res_val = world_bank('ESP', ref_year, 'NY.GDP.MKTP.CD')

        ref_year = 1960
        ref_val = 12072126075.397
        self.assertEqual(res_year, ref_year)
        self.assertEqual(res_val, ref_val)

class TestWealth2GDP(unittest.TestCase):
    """Test Wealth to GDP factor extraction"""
    def test_nfw_SUR_pass(self):
        """Test non-financial wealth-to-gdp factor with Suriname."""
        res_year, res_val = wealth2gdp('SUR')

        ref_year = 2016
        ref_val = 0.73656
        self.assertEqual(res_year, ref_year)
        self.assertEqual(res_val, ref_val)

    def test_nfw_BEL_pass(self):
        """Test total wealth-to-gdp factor with Belgium."""
        res_year, res_val = wealth2gdp('BEL', False)

        ref_year = 2016
        ref_val = 4.88758
        self.assertEqual(res_year, ref_year)
        self.assertEqual(res_val, ref_val)

    def test_nfw_LBY_pass(self):
        """Test missing factor with Libya."""
        _, res_val = wealth2gdp('LBY')

        self.assertTrue(np.isnan(res_val))

class TestWBWealthAccount(unittest.TestCase):
    """Test Wealth Indicator extraction from World Bank provided CSV"""
    def test_pca_DEU_2010_pass(self):
        """Test Processed Capital value Germany 2010."""
        ref_year = 2010
        cntry_iso = 'DEU'
        res_year, res_val, q = world_bank_wealth_account(cntry_iso, ref_year, no_land=0)
        res_year_noland, res_val_noland, q = world_bank_wealth_account(cntry_iso, ref_year,
                                                                       no_land=1)
        ref_val = [17675048450284.9, 19767982562092.2]  # second value as updated by worldbank on
                                                        # October 27 2021
        ref_val_noland = [14254071330874.9, 15941921421042.1]  # dito
        self.assertEqual(res_year, ref_year)
        self.assertEqual(q, 1)
        self.assertIn(res_val, ref_val)
        self.assertEqual(res_year_noland, ref_year)
        self.assertIn(res_val_noland, ref_val_noland)

    def test_pca_CHE_2008_pass(self):
        """Test Prcoessed Capital per capita Switzerland 2008 (interp.)."""
        ref_year = 2008
        cntry_iso = 'CHE'
        var_name = 'NW.PCA.PC'
        res_year, res_val, _ = world_bank_wealth_account(cntry_iso, ref_year,
                                                         variable_name=var_name, no_land=0)
        ref_val = [328398.7, 369081.0]  # second value as updated by worldbank on October 27 2021
        self.assertEqual(res_year, ref_year)
        self.assertIn(res_val, ref_val)

    def test_tow_IND_1985_pass(self):
        """Test Total Wealth value India 1985 (outside year range)."""
        ref_year = 1985
        cntry_iso = 'IND'
        var_name = 'NW.TOW.TO'
        res_year, res_val, _ = world_bank_wealth_account(cntry_iso, ref_year,
                                                         variable_name=var_name)
        ref_val = [5415188681942.5, 5861193808788.2]  # second value as updated by worldbank on 
                                                      # October 27 2021
        self.assertEqual(res_year, ref_year)
        self.assertIn(res_val, ref_val)

    def test_pca_CUB_2015_pass(self):
        """Test Processed Capital value Cuba 2015 (missing value)."""
        ref_year = 2015
        cntry_iso = 'CUB'
        res_year, res_val, q = world_bank_wealth_account(cntry_iso, ref_year, no_land=1)
        ref_val = 108675762920.0
        self.assertEqual(q, 0)
        self.assertEqual(res_year, ref_year)
        self.assertEqual(res_val, ref_val)

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestNetpresValue)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestWBData))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestWealth2GDP))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestWBWealthAccount))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test tc_tracks module.
"""

import unittest
import xarray as xr
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point, LineString, MultiLineString

import climada.hazard.tc_tracks as tc
from climada import CONFIG
from climada.util import ureg
from climada.util.constants import TC_ANDREW_FL
import climada.util.coordinates as u_coord
from climada.entity import Exposures
from datetime import datetime as dt

DATA_DIR = CONFIG.hazard.test_data.dir()
TEST_TRACK = DATA_DIR.joinpath("trac_brb_test.csv")
TEST_TRACK_SHORT = DATA_DIR.joinpath("trac_short_test.csv")
TEST_RAW_TRACK = DATA_DIR.joinpath('Storm.2016075S11087.ibtracs_all.v03r10.csv')
TEST_TRACK_GETTELMAN = DATA_DIR.joinpath('gettelman_test_tracks.nc')
TEST_TRACK_EMANUEL = DATA_DIR.joinpath('emanuel_test_tracks.mat')
TEST_TRACK_EMANUEL_CORR = DATA_DIR.joinpath('temp_mpircp85cal_full.mat')
TEST_TRACK_CHAZ = DATA_DIR.joinpath('chaz_test_tracks.nc')
TEST_TRACK_STORM = DATA_DIR.joinpath('storm_test_tracks.txt')
TEST_TRACKS_ANTIMERIDIAN = DATA_DIR.joinpath('tracks-antimeridian')


class TestIbtracs(unittest.TestCase):
    """Test reading and model of TC from IBTrACS files"""

    def test_raw_ibtracs_empty_pass(self):
        """Test reading empty TC from IBTrACS files"""
        tc_track = tc.TCTracks.from_ibtracs_netcdf(
            provider='usa', storm_id='1988234N13299')
        self.assertEqual(tc_track.size, 0)
        self.assertEqual(tc_track.get_track(), [])

    def test_raw_ibtracs_invalid_pass(self):
        """Test reading invalid/non-existing TC from IBTrACS files"""
        with self.assertRaises(ValueError) as cm:
            tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id='INVALID')
        self.assertIn("IDs are invalid", str(cm.exception))
        self.assertIn("INVALID", str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id='1988234N13298')
        self.assertIn("IDs are not in IBTrACS", str(cm.exception))
        self.assertIn("1988234N13298", str(cm.exception))

    def test_penv_rmax_penv_pass(self):
        """from_ibtracs_netcdf"""
        tc_track = tc.TCTracks.from_ibtracs_netcdf(provider='usa', storm_id='1992230N11325')
        penv_ref = np.ones(97) * 1010
        penv_ref[26:36] = [1011, 1012, 1013, 1014, 1015, 1014, 1014, 1014, 1014, 1012]

        self.assertTrue(np.allclose(
            tc_track.get_track().environmental_pressure.values, penv_ref))
        self.assertTrue(np.allclose(
            tc_track.get_track().radius_max_wind.values, np.zeros(97)))

    def test_ibtracs_raw_pass(self):
        """Read a tropical cyclone."""
        # read without specified provider or estimation of missing values
        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id='2017242N16333')
        track_ds = tc_track.get_track()
        self.assertEqual(len(tc_track.data), 1)
        self.assertEqual(track_ds.time.dt.year.values[0], 2017)
        self.assertEqual(track_ds.time.dt.month.values[0], 8)
        self.assertEqual(track_ds.time.dt.day.values[0], 30)
        self.assertEqual(track_ds.time.dt.hour.values[0], 0)
        self.assertAlmostEqual(track_ds.lat.values[0], 16.1, places=5)
        self.assertAlmostEqual(track_ds.lon.values[0], -26.9, places=5)
        self.assertAlmostEqual(track_ds.max_sustained_wind.values[0], 30)
        self.assertAlmostEqual(track_ds.central_pressure.values[0], 1008)
        self.assertAlmostEqual(track_ds.environmental_pressure.values[0], 1012)
        self.assertAlmostEqual(track_ds.radius_max_wind.values[0], 60)
        self.assertEqual(track_ds.time.size, 123)

        self.assertAlmostEqual(track_ds.lat.values[-1], 36.8, places=5)
        self.assertAlmostEqual(track_ds.lon.values[-1], -90.1, places=4)
        self.assertAlmostEqual(track_ds.central_pressure.values[-1], 1005)
        self.assertAlmostEqual(track_ds.max_sustained_wind.values[-1], 15)
        self.assertAlmostEqual(track_ds.environmental_pressure.values[-1], 1008)
        self.assertAlmostEqual(track_ds.radius_max_wind.values[-1], 60)

        self.assertFalse(np.isnan(track_ds.radius_max_wind.values).any())
        self.assertFalse(np.isnan(track_ds.environmental_pressure.values).any())
        self.assertFalse(np.isnan(track_ds.max_sustained_wind.values).any())
        self.assertFalse(np.isnan(track_ds.central_pressure.values).any())
        self.assertFalse(np.isnan(track_ds.lat.values).any())
        self.assertFalse(np.isnan(track_ds.lon.values).any())

        np.testing.assert_array_equal(track_ds.basin, 'NA')
        self.assertEqual(track_ds.max_sustained_wind_unit, 'kn')
        self.assertEqual(track_ds.central_pressure_unit, 'mb')
        self.assertEqual(track_ds.sid, '2017242N16333')
        self.assertEqual(track_ds.name, 'IRMA')
        self.assertEqual(track_ds.orig_event_flag, True)
        self.assertEqual(track_ds.data_provider,
                         'ibtracs_mixed:lat(official_3h),lon(official_3h),wind(official_3h),'
                         'pres(official_3h),rmw(official_3h),poci(official_3h),roci(official_3h)')
        self.assertEqual(track_ds.category, 5)

    def test_ibtracs_with_provider(self):
        """Read a tropical cyclone with and without explicit provider."""
        storm_id = '2012152N12130'
        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id=storm_id, provider='usa')
        track_ds = tc_track.get_track()
        self.assertEqual(track_ds.time.size, 51)
        self.assertEqual(track_ds.data_provider, 'ibtracs_usa')
        self.assertAlmostEqual(track_ds.lat.values[50], 34.3, places=5)
        self.assertAlmostEqual(track_ds.central_pressure.values[50], 989, places=5)
        self.assertAlmostEqual(track_ds.radius_max_wind.values[46], 20, places=5)

        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id=storm_id)
        track_ds = tc_track.get_track()
        self.assertEqual(track_ds.time.size, 35)
        self.assertEqual(track_ds.data_provider,
                         'ibtracs_mixed:lat(official_3h),lon(official_3h),wind(official_3h),'
                         'pres(official_3h),rmw(usa),poci(usa),roci(usa)')
        self.assertAlmostEqual(track_ds.lat.values[-1], 31.40, places=5)
        self.assertAlmostEqual(track_ds.central_pressure.values[-1], 980, places=5)

    def test_ibtracs_antimeridian(self):
        """Read a track that crosses the antimeridian and make sure that lon is consistent"""
        storm_id = '2013224N12220'

        # the officially responsible agencies 'usa' and 'tokyo' use different signs in lon, but we
        # have to `estimate_missing` because both have gaps in reported values
        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id=storm_id, provider=['official_3h'],
                                     estimate_missing=True)
        track_ds = tc_track.get_track()
        np.testing.assert_array_less(0, track_ds.lon)

    def test_ibtracs_estimate_missing(self):
        """Read a tropical cyclone and estimate missing values."""
        storm_id = '2012152N12130'

        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id=storm_id, estimate_missing=True)
        track_ds = tc_track.get_track()
        # less time steps are discarded, leading to a larger total size
        self.assertEqual(track_ds.time.size, 99)
        self.assertEqual(track_ds.data_provider,
                         'ibtracs_mixed:lat(official_3h),lon(official_3h),wind(official_3h),'
                         'pres(official_3h),rmw(usa),poci(usa),roci(usa)')
        self.assertAlmostEqual(track_ds.lat.values[44], 33.30, places=5)
        self.assertAlmostEqual(track_ds.central_pressure.values[44], 976, places=5)
        self.assertAlmostEqual(track_ds.central_pressure.values[42], 980, places=5)
        # the wind speed at position 44 is missing in the original data
        self.assertAlmostEqual(track_ds.max_sustained_wind.values[44], 58, places=0)
        self.assertAlmostEqual(track_ds.radius_oci.values[40], 160, places=0)
        # after position 42, ROCI is missing in the original data
        self.assertAlmostEqual(track_ds.radius_oci.values[42], 200, places=-1)
        self.assertAlmostEqual(track_ds.radius_oci.values[85], 165, places=-1)
        self.assertAlmostEqual(track_ds.radius_oci.values[95], 155, places=-1)

    def test_ibtracs_official(self):
        """Read a tropical cyclone, only officially reported values."""
        storm_id = '2012152N12130'

        tc_track = tc.TCTracks.from_ibtracs_netcdf(
            storm_id=storm_id, interpolate_missing=False, provider='official')
        track_ds = tc_track.get_track()
        self.assertEqual(track_ds.time.size, 21)
        self.assertEqual(track_ds.data_provider, 'ibtracs_official')
        self.assertAlmostEqual(track_ds.lon.values[19], 137.6, places=4)
        self.assertAlmostEqual(track_ds.central_pressure.values[19], 980, places=5)
        np.testing.assert_array_equal(track_ds.radius_max_wind.values, 0)

    def test_ibtracs_scale_wind(self):
        """Read a tropical cyclone and scale wind speed according to agency."""
        storm_id = '2012152N12130'

        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id=storm_id, rescale_windspeeds=True)
        track_ds = tc_track.get_track()
        self.assertAlmostEqual(track_ds.max_sustained_wind.values[34], (55 - 23.3) / 0.6, places=5)

        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id=storm_id, rescale_windspeeds=False)
        track_ds = tc_track.get_track()
        self.assertAlmostEqual(track_ds.max_sustained_wind.values[34], 55, places=5)

    def test_ibtracs_interpolate_missing(self):
        """Read a tropical cyclone with and without interpolating missing values."""
        storm_id = '2010066S19050'

        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id=storm_id, interpolate_missing=False)
        track_ds = tc_track.get_track()
        self.assertEqual(track_ds.time.size, 50)
        self.assertAlmostEqual(track_ds.central_pressure.values[30], 992, places=5)
        self.assertAlmostEqual(track_ds.central_pressure.values[31], 1006, places=5)

        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id=storm_id, interpolate_missing=True)
        track_ds = tc_track.get_track()
        self.assertEqual(track_ds.time.size, 65)
        self.assertAlmostEqual(track_ds.central_pressure.values[30], 992, places=5)
        self.assertAlmostEqual(track_ds.central_pressure.values[38], 999, places=5)
        self.assertAlmostEqual(track_ds.central_pressure.values[46], 1006, places=5)

    def test_ibtracs_range(self):
        """Read several TCs."""
        tc_track = tc.TCTracks.from_ibtracs_netcdf(year_range=(2100, 2150))
        self.assertEqual(tc_track.size, 0)

        tc_track = tc.TCTracks.from_ibtracs_netcdf(
            provider='usa', storm_id=None, year_range=(1915, 1916), basin='WP')
        self.assertEqual(tc_track.size, 0)

        tc_track = tc.TCTracks.from_ibtracs_netcdf(
            provider='usa', year_range=(1993, 1994), basin='EP', estimate_missing=False)
        self.assertEqual(tc_track.size, 33)

        tc_track = tc.TCTracks.from_ibtracs_netcdf(
            provider='usa', year_range=(1993, 1994), basin='EP', estimate_missing=True)
        self.assertEqual(tc_track.size, 45)

    def test_ibtracs_correct_pass(self):
        """Check estimate_missing option"""
        tc_try = tc.TCTracks.from_ibtracs_netcdf(
            provider='usa', storm_id='1982267N25289', estimate_missing=True)
        self.assertAlmostEqual(tc_try.data[0].central_pressure.values[0], 1013, places=0)
        self.assertAlmostEqual(tc_try.data[0].central_pressure.values[5], 1008, places=0)
        self.assertAlmostEqual(tc_try.data[0].central_pressure.values[-1], 1012, places=0)

    def test_ibtracs_with_basin(self):
        """Filter TCs by (genesis) basin."""
        # South Atlantic (not usually a TC location at all)
        tc_track = tc.TCTracks.from_ibtracs_netcdf(basin="SA")
        self.assertEqual(tc_track.size, 3)

        # the basin is not necessarily the genesis basin
        tc_track = tc.TCTracks.from_ibtracs_netcdf(
            year_range=(1995, 1995), basin="SP", estimate_missing=True)
        self.assertEqual(tc_track.size, 6)
        self.assertEqual(tc_track.data[0].basin[0], 'SP')
        self.assertEqual(tc_track.data[5].basin[0], 'SI')

        # genesis in NI
        tc_track = tc.TCTracks.from_ibtracs_netcdf(
            year_range=(1994, 1994), genesis_basin="NI", estimate_missing=True)
        self.assertEqual(tc_track.size, 5)
        for tr in tc_track.data:
            self.assertEqual(tr.basin[0], "NI")

        # genesis in EP, but crosses WP at some point
        tc_track = tc.TCTracks.from_ibtracs_netcdf(
            year_range=(2002, 2003), basin="WP", genesis_basin="EP")
        self.assertEqual(tc_track.size, 3)
        for tr in tc_track.data:
            self.assertEqual(tr.basin[0], "EP")
            self.assertIn("WP", tr.basin)

    def test_ibtracs_discard_single_points(self):
        """Check discard_single_points option"""
        passed = False
        for year in range(1863, 1981):
            tc_track_singlept = tc.TCTracks.from_ibtracs_netcdf(
                provider='usa', year_range=(year,year), discard_single_points=False)
            n_singlepts = np.sum([x.time.size == 1 for x in tc_track_singlept.data])
            if n_singlepts > 0:
                tc_track = tc.TCTracks.from_ibtracs_netcdf(provider='usa', year_range=(year,year))
                if tc_track.size == tc_track_singlept.size - n_singlepts:
                    passed = True
                    break
        self.assertTrue(passed)

class TestIO(unittest.TestCase):
    """Test reading of tracks from files of different formats"""
    def test_netcdf_io(self):
        """Test writting and reading netcdf4 TCTracks instances"""
        path = DATA_DIR.joinpath("tc_tracks_nc")
        path.mkdir(exist_ok=True)
        tc_track = tc.TCTracks.from_ibtracs_netcdf(
            provider='usa', storm_id='1988234N13299', estimate_missing=True)
        tc_track.write_netcdf(str(path))

        tc_read = tc.TCTracks.from_netcdf(str(path))
        self.assertEqual(tc_track.get_track().sid, tc_read.get_track().sid)

    def test_read_legacy_netcdf(self):
        """Test reading from NetCDF files with legacy basin attributes"""
        # test data set with two tracks:
        # * 1980052S16155: crosses the antimeridian
        # * 2018079S09162: close, but doesn't cross antimeridian; has self-intersections
        anti_track = tc.TCTracks.from_netcdf(TEST_TRACKS_ANTIMERIDIAN)

        for tr in anti_track.data:
            self.assertEqual(tr.basin.shape, tr.time.shape)
            np.testing.assert_array_equal(tr.basin, "SP")

    def test_hdf5_io(self):
        """Test writting and reading hdf5 TCTracks instances"""
        path = DATA_DIR.joinpath("tc_tracks.h5")
        tc_track = tc.TCTracks.from_ibtracs_netcdf(
            provider='usa', year_range=(1993, 1994), basin='EP', estimate_missing=True)
        tc_track.write_hdf5(path)
        tc_read = tc.TCTracks.from_hdf5(path)
        path.unlink()

        self.assertEqual(len(tc_track.data), len(tc_read.data))
        for tr, tr_read in zip(tc_track.data, tc_read.data):
            self.assertEqual(set(tr.attrs.keys()), set(tr_read.attrs.keys()))
            self.assertEqual(set(tr.variables), set(tr_read.variables))
            self.assertEqual(set(tr.coords), set(tr_read.coords))
            for key in tr.attrs.keys():
                self.assertEqual(tr.attrs[key], tr_read.attrs[key])
            for v in tr.variables:
                self.assertEqual(tr[v].dtype, tr_read[v].dtype)
                np.testing.assert_array_equal(tr[v].values, tr_read[v].values)
            self.assertEqual(tr.sid, tr_read.sid)

    def test_from_processed_ibtracs_csv(self):
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK)

        self.assertEqual(tc_track.data[0].time.size, 38)
        self.assertEqual(tc_track.data[0].lon[11], -39.60)
        self.assertEqual(tc_track.data[0].lat[23], 14.10)
        self.assertEqual(tc_track.data[0].time_step[7], 6)
        self.assertEqual(np.max(tc_track.data[0].radius_max_wind), 0)
        self.assertEqual(np.min(tc_track.data[0].radius_max_wind), 0)
        self.assertEqual(tc_track.data[0].max_sustained_wind[21], 55)
        self.assertAlmostEqual(tc_track.data[0].central_pressure.values[29], 976, places=0)
        self.assertEqual(np.max(tc_track.data[0].environmental_pressure), 1010)
        self.assertEqual(np.min(tc_track.data[0].environmental_pressure), 1010)
        self.assertEqual(tc_track.data[0].time.dt.year[13], 1951)
        self.assertEqual(tc_track.data[0].time.dt.month[26], 9)
        self.assertEqual(tc_track.data[0].time.dt.day[7], 29)
        self.assertEqual(tc_track.data[0].max_sustained_wind_unit, 'kn')
        self.assertEqual(tc_track.data[0].central_pressure_unit, 'mb')
        self.assertEqual(tc_track.data[0].orig_event_flag, 1)
        self.assertEqual(tc_track.data[0].name, '1951239N12334')
        self.assertEqual(tc_track.data[0].sid, '1951239N12334')
        self.assertEqual(tc_track.data[0].id_no, 1951239012334)
        self.assertEqual(tc_track.data[0].data_provider, 'hurdat_atl')
        np.testing.assert_array_equal(tc_track.data[0].basin, 'NA')
        self.assertEqual(tc_track.data[0].id_no, 1951239012334)
        self.assertEqual(tc_track.data[0].category, 1)

    def test_from_simulations_emanuel(self):
        tc_track = tc.TCTracks.from_simulations_emanuel(TEST_TRACK_EMANUEL, hemisphere='N')
        self.assertEqual(len(tc_track.data), 4)
        self.assertEqual(tc_track.data[0].time.size, 93)
        self.assertEqual(tc_track.data[0].lon[11], -115.57)
        self.assertEqual(tc_track.data[0].lat[23], 10.758)
        self.assertEqual(tc_track.data[0].time_step[7], 2)
        self.assertAlmostEqual(tc_track.data[0].radius_max_wind[15], 44.27645788336934)
        self.assertEqual(tc_track.data[0].max_sustained_wind[21], 27.1)
        self.assertEqual(tc_track.data[0].central_pressure[29], 995.31)
        self.assertTrue(np.all(tc_track.data[0].environmental_pressure == 1010))
        self.assertTrue(np.all(tc_track.data[0].time.dt.year == 1950))
        self.assertEqual(tc_track.data[0].time.dt.month[26], 10)
        self.assertEqual(tc_track.data[0].time.dt.day[7], 26)
        self.assertEqual(tc_track.data[0].max_sustained_wind_unit, 'kn')
        self.assertEqual(tc_track.data[0].central_pressure_unit, 'mb')
        self.assertEqual(tc_track.data[0].sid, '1')
        self.assertEqual(tc_track.data[0].name, '1')
        self.assertEqual(tc_track.data[0].basin.dtype, '<U2')
        self.assertTrue(np.all([np.all(d.basin == 'N') for d in tc_track.data]))
        self.assertEqual(tc_track.data[0].category, 3)

        tc_track = tc.TCTracks.from_simulations_emanuel(TEST_TRACK_EMANUEL_CORR, hemisphere='S')
        self.assertEqual(len(tc_track.data), 2)
        self.assertTrue(np.all([np.all(d.basin == 'S') for d in tc_track.data]))
        self.assertEqual(tc_track.data[0].radius_max_wind[15], 102.49460043196545)
        self.assertEqual(tc_track.data[0].time.dt.month[343], 2)
        self.assertEqual(tc_track.data[0].time.dt.day[343], 28)
        self.assertEqual(tc_track.data[0].time.dt.month[344], 3)
        self.assertEqual(tc_track.data[0].time.dt.day[344], 1)
        self.assertEqual(tc_track.data[1].time.dt.year[0], 2009)
        self.assertEqual(tc_track.data[1].time.dt.year[256], 2009)
        self.assertEqual(tc_track.data[1].time.dt.year[257], 2010)
        self.assertEqual(tc_track.data[1].time.dt.year[-1], 2010)

        tc_track = tc.TCTracks.from_simulations_emanuel(TEST_TRACK_EMANUEL_CORR)
        self.assertEqual(len(tc_track.data), 5)
        self.assertTrue(np.all([np.all(d.basin == 'GB') for d in tc_track.data]))

    def test_read_gettelman(self):
        """Test reading and model of TC from Gettelman track files"""
        tc_track_G = tc.TCTracks.from_gettelman(TEST_TRACK_GETTELMAN)
        self.assertEqual(tc_track_G.data[0].time.size, 29)
        self.assertEqual(tc_track_G.data[0].lon[11], 60.0)
        self.assertEqual(tc_track_G.data[0].lat[23], 10.20860481262207)
        self.assertEqual(tc_track_G.data[0].time_step[7], 3.)
        self.assertEqual(np.max(tc_track_G.data[0].radius_max_wind), 65)
        self.assertEqual(np.min(tc_track_G.data[0].radius_max_wind), 65)
        self.assertEqual(tc_track_G.data[0].max_sustained_wind[21], 39.91877223718089)
        self.assertEqual(tc_track_G.data[0].central_pressure[27], 1005.969482421875)
        self.assertEqual(np.max(tc_track_G.data[0].environmental_pressure), 1015)
        self.assertEqual(np.min(tc_track_G.data[0].environmental_pressure), 1015)
        self.assertEqual(tc_track_G.data[0].maximum_precipitation[14], 219.10108947753906)
        self.assertEqual(tc_track_G.data[0].average_precipitation[12], 101.43893432617188)
        self.assertEqual(tc_track_G.data[0].time.dt.year[13], 1979)
        self.assertEqual(tc_track_G.data[0].time.dt.month[26], 1)
        self.assertEqual(tc_track_G.data[0].time.dt.day[7], 2)
        self.assertEqual(tc_track_G.data[0].max_sustained_wind_unit, 'kn')
        self.assertEqual(tc_track_G.data[0].central_pressure_unit, 'mb')
        self.assertEqual(tc_track_G.data[0].sid, '0')
        self.assertEqual(tc_track_G.data[0].name, '0')
        self.assertEqual(tc_track_G.data[0].basin.dtype, '<U2')
        np.testing.assert_array_equal(tc_track_G.data[0].basin, 'NI')
        self.assertEqual(tc_track_G.data[0].category, 0)

    def test_from_simulations_chaz(self):
        """Test reading NetCDF output from CHAZ simulations"""
        tc_track = tc.TCTracks.from_simulations_chaz(TEST_TRACK_CHAZ)
        self.assertEqual(len(tc_track.data), 13)
        self.assertEqual(tc_track.data[0].time.size, 5)
        self.assertEqual(tc_track.data[0].lon[3], 74.1388328911036)
        self.assertEqual(tc_track.data[0].lat[4], -9.813585651475156)
        self.assertEqual(tc_track.data[0].time_step[3], 6)
        self.assertEqual(tc_track.data[0].max_sustained_wind[2], 20.188325232226354)
        self.assertAlmostEqual(tc_track.data[0].central_pressure.values[1], 1008, places=0)
        self.assertTrue(np.all(tc_track.data[0].time.dt.year == 1991))
        self.assertEqual(tc_track.data[0].time.dt.month[2], 1)
        self.assertEqual(tc_track.data[0].time.dt.day[3], 15)
        self.assertEqual(tc_track.data[0].max_sustained_wind_unit, 'kn')
        self.assertEqual(tc_track.data[0].central_pressure_unit, 'mb')
        self.assertEqual(tc_track.data[0].sid, 'chaz_test_tracks.nc-1-0')
        self.assertEqual(tc_track.data[0].name, 'chaz_test_tracks.nc-1-0')
        self.assertTrue(np.all([np.all(d.basin == 'GB') for d in tc_track.data]))
        self.assertEqual(tc_track.data[4].category, 0)
        self.assertEqual(tc_track.data[3].category, -1)

        tc_track = tc.TCTracks.from_simulations_chaz(TEST_TRACK_CHAZ, year_range=(1990, 1991))
        self.assertEqual(len(tc_track.data), 3)

        tc_track = tc.TCTracks.from_simulations_chaz(TEST_TRACK_CHAZ, year_range=(1950, 1955))
        self.assertEqual(len(tc_track.data), 0)

        tc_track = tc.TCTracks.from_simulations_chaz(TEST_TRACK_CHAZ, ensemble_nums=[0, 2])
        self.assertEqual(len(tc_track.data), 9)

    def test_from_simulations_storm(self):
        """Test reading NetCDF output from STORM simulations"""
        tc_track = tc.TCTracks.from_simulations_storm(TEST_TRACK_STORM)
        self.assertEqual(len(tc_track.data), 6)
        self.assertEqual(tc_track.data[0].time.size, 15)
        self.assertEqual(tc_track.data[0].lon[3], 245.3)
        self.assertEqual(tc_track.data[0].lat[4], 11.9)
        self.assertEqual(tc_track.data[0].time_step[3], 3)
        self.assertEqual(tc_track.data[0].max_sustained_wind[2], 42.19026114274495)
        self.assertEqual(tc_track.data[0].radius_max_wind[5], 19.07407454551836)
        self.assertEqual(tc_track.data[0].central_pressure[1], 999.4)
        self.assertTrue(np.all(tc_track.data[0].time.dt.year == 1980))
        self.assertEqual(tc_track.data[0].time.dt.month[2].item(), 6)
        self.assertEqual(tc_track.data[0].time.dt.day[3].item(), 1)
        self.assertEqual(tc_track.data[0].max_sustained_wind_unit, 'kn')
        self.assertEqual(tc_track.data[0].central_pressure_unit, 'mb')
        self.assertEqual(tc_track.data[0].sid, 'storm_test_tracks.txt-0-0')
        self.assertEqual(tc_track.data[0].name, 'storm_test_tracks.txt-0-0')
        self.assertTrue(np.all([np.all(d.basin == 'EP') for d in tc_track.data]))
        self.assertEqual(tc_track.data[4].category, 0)
        self.assertEqual(tc_track.data[3].category, 1)

        tc_track = tc.TCTracks.from_simulations_storm(TEST_TRACK_STORM, years=[0, 2])
        self.assertEqual(len(tc_track.data), 4)

        tc_track = tc.TCTracks.from_simulations_storm(TEST_TRACK_STORM, years=[7])
        self.assertEqual(len(tc_track.data), 0)

    def test_to_geodataframe_points(self):
        """Conversion of TCTracks to GeoDataFrame using Points."""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK)

        gdf_points = tc_track.to_geodataframe(as_points=True)
        self.assertIsInstance(gdf_points.unary_union.bounds, tuple)
        self.assertEqual(gdf_points.shape[0], len(tc_track.data[0].time))
        self.assertEqual(gdf_points.shape[1], len(tc_track.data[0].variables)+len(tc_track.data[0].attrs)-1)
        self.assertAlmostEqual(gdf_points.buffer(3).unary_union.area, 348.79972062947854)
        self.assertIsInstance(gdf_points.iloc[0].time, pd._libs.tslibs.timestamps.Timestamp)

    def test_to_geodataframe_line(self):
        """Conversion of TCTracks to GeoDataFrame using LineStrings."""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK)

        gdf_line = tc_track.to_geodataframe()
        self.assertEqual(gdf_line.shape[0], 1)
        self.assertAlmostEqual(gdf_line.geometry[0].length, 54.0634224372971)
        self.assertIsInstance(gdf_line.bounds.minx, pd.core.series.Series)

        # test data set with two tracks:
        # * 1980052S16155: crosses the antimeridian
        # * 2018079S09162: close, but doesn't cross antimeridian; has self-intersections
        anti_track = tc.TCTracks.from_netcdf(TEST_TRACKS_ANTIMERIDIAN)

        split = anti_track.to_geodataframe(split_lines_antimeridian=True)
        split.set_index('sid', inplace=True)
        self.assertIsInstance(split.loc['1980052S16155'].geometry, MultiLineString)
        self.assertIsInstance(split.loc['2018079S09162'].geometry, LineString)
        self.assertEqual(len(split.loc['1980052S16155'].geometry), 8)
        self.assertFalse(split.loc['2018079S09162'].geometry.is_simple)

        nosplit = anti_track.to_geodataframe(split_lines_antimeridian=False)
        nosplit.set_index('sid', inplace=True)
        self.assertIsInstance(nosplit.loc['1980052S16155'].geometry, LineString)
        self.assertIsInstance(nosplit.loc['2018079S09162'].geometry, LineString)
        self.assertFalse(nosplit.loc['2018079S09162'].geometry.is_simple)

        np.testing.assert_array_almost_equal(
            split.loc['1980052S16155'].geometry.bounds, (-180, -38.583244, 180, -17))
        np.testing.assert_array_almost_equal(
            nosplit.loc['1980052S16155'].geometry.bounds, (150.800003, -38.583244, 185, -17))
        np.testing.assert_array_almost_equal(
            split.loc['2018079S09162'].geometry.bounds, (148.600006, -22.41, 162.399994, -13.4))
        np.testing.assert_array_almost_equal(
            nosplit.loc['2018079S09162'].geometry.bounds, (148.600006, -22.41, 162.399994, -13.4))

class TestFuncs(unittest.TestCase):
    """Test functions over TC tracks"""

    def test_get_track_pass(self):
        """Test get_track."""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK_SHORT)
        self.assertIsInstance(tc_track.get_track(), xr.Dataset)
        self.assertIsInstance(tc_track.get_track('1951239N12334'), xr.Dataset)

        tc_track_bis = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK_SHORT)
        tc_track.append(tc_track_bis)
        self.assertIsInstance(tc_track.get_track(), list)
        self.assertIsInstance(tc_track.get_track('1951239N12334'), xr.Dataset)

    def test_subset(self):
        """Test subset."""
        storms = ['1988169N14259', '2002073S16161', '2002143S07157']
        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id=storms)
        self.assertEqual(tc_track.subset({'basin': 'SP'}).size, 2)

    def test_get_extent(self):
        """Test extent/bounds attributes."""
        storms = ['1988169N14259', '2002073S16161', '2002143S07157']
        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id=storms, provider=["usa", "bom"])
        bounds = (153.585022, -23.200001, 258.714996, 17.514986)
        extent = (bounds[0], bounds[2], bounds[1], bounds[3])
        bounds_buf = (153.485022, -23.300001, 258.814996, 17.614986)
        np.testing.assert_array_almost_equal(tc_track.bounds, bounds)
        np.testing.assert_array_almost_equal(tc_track.get_bounds(deg_buffer=0.1), bounds_buf)
        np.testing.assert_array_almost_equal(tc_track.extent, extent)

    def test_generate_centroids(self):
        """Test centroids generation feature."""
        storms = ['1988169N14259', '2002073S16161', '2002143S07157']
        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id=storms, provider=["usa", "bom"])
        cen = tc_track.generate_centroids(10, 1)
        cen_bounds = (157.585022, -19.200001, 257.585022, 10.799999)
        self.assertEqual(cen.size, 44)
        self.assertEqual(np.unique(cen.lat).size, 4)
        self.assertEqual(np.unique(cen.lon).size, 11)
        np.testing.assert_array_equal(np.diff(np.unique(cen.lat)), 10)
        np.testing.assert_array_equal(np.diff(np.unique(cen.lon)), 10)
        np.testing.assert_array_almost_equal(cen.total_bounds, cen_bounds)

    def test_interp_track_pass(self):
        """Interpolate track to min_time_step. Compare to MATLAB reference."""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK)
        tc_track.equal_timestep(time_step_h=1)

        self.assertEqual(tc_track.data[0].time.size, 223)
        self.assertAlmostEqual(tc_track.data[0].lon.values[11], -27.426151640151684)
        self.assertAlmostEqual(float(tc_track.data[0].lat[23]), 12.300006169591480)
        self.assertEqual(tc_track.data[0].time_step[7], 1)
        self.assertEqual(np.max(tc_track.data[0].radius_max_wind), 0)
        self.assertEqual(np.min(tc_track.data[0].radius_max_wind), 0)
        self.assertEqual(tc_track.data[0].max_sustained_wind[21], 25)
        self.assertTrue(np.isfinite(tc_track.data[0].central_pressure.values).all())
        self.assertAlmostEqual(tc_track.data[0].central_pressure.values[29], 1008, places=0)
        self.assertEqual(np.max(tc_track.data[0].environmental_pressure), 1010)
        self.assertEqual(np.min(tc_track.data[0].environmental_pressure), 1010)
        self.assertEqual(tc_track.data[0]['time.year'][13], 1951)
        self.assertEqual(tc_track.data[0]['time.month'][26], 8)
        self.assertEqual(tc_track.data[0]['time.day'][7], 27)
        self.assertEqual(tc_track.data[0].max_sustained_wind_unit, 'kn')
        self.assertEqual(tc_track.data[0].central_pressure_unit, 'mb')
        self.assertEqual(tc_track.data[0].orig_event_flag, 1)
        self.assertEqual(tc_track.data[0].name, '1951239N12334')
        self.assertEqual(tc_track.data[0].data_provider, 'hurdat_atl')
        np.testing.assert_array_equal(tc_track.data[0].basin, 'NA')
        self.assertEqual(tc_track.data[0].id_no, 1951239012334)
        self.assertEqual(tc_track.data[0].category, 1)

        # test some "generic floats"
        for time_step_h in [0.6663545049172093, 2.509374054925788, 8.175754471661111]:
            # artifically create data that doesn't start at full hour
            for loffset in [0, 22, 30]:
                tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK)
                tc_track.data[0].time.values[:] += np.timedelta64(loffset, "m")
                tc_track.equal_timestep(time_step_h=time_step_h)
                np.testing.assert_array_equal(tc_track.data[0].time_step, time_step_h)
                self.assertTrue(np.isfinite(tc_track.data[0].central_pressure.values).all())

        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK)
        tc_track.equal_timestep(time_step_h=0.16667)

        self.assertEqual(tc_track.data[0].time.size, 1332)
        self.assertTrue(np.all(tc_track.data[0].time_step == 0.16667))
        self.assertTrue(np.isfinite(tc_track.data[0].central_pressure.values).all())
        self.assertAlmostEqual(tc_track.data[0].lon.values[65], -27.397636528537127)

        for time_step_h in [0, -0.5, -1]:
            tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK)
            msg = f"time_step_h is not a positive number: {time_step_h}"
            with self.assertRaises(ValueError, msg=msg) as _cm:
                tc_track.equal_timestep(time_step_h=time_step_h)

    def test_interp_origin_pass(self):
        """Interpolate track to min_time_step crossing lat origin"""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK)
        tc_track.data[0].lon.values = np.array([
            167.207761, 168.1, 168.936535, 169.728947, 170.5, 171.257176,
            171.946822, 172.5, 172.871797, 173.113396, 173.3, 173.496375,
            173.725522, 174., 174.331591, 174.728961, 175.2, 175.747632,
            176.354929, 177., 177.66677, 178.362433, 179.1, 179.885288,
            -179.304661, -178.5, -177.726442, -176.991938, -176.3, -175.653595,
            -175.053513, -174.5, -173.992511, -173.527342, -173.1, -172.705991,
            -172.340823, -172.
        ])
        tc_track.data[0].lat.values = np.array([
            40.196053, 40.6, 40.930215, 41.215674, 41.5, 41.816354, 42.156065,
            42.5, 42.833998, 43.16377, 43.5, 43.847656, 44.188854, 44.5,
            44.764269, 44.991925, 45.2, 45.402675, 45.602707, 45.8, 45.995402,
            46.193543, 46.4, 46.615718, 46.82312, 47., 47.130616, 47.225088,
            47.3, 47.369224, 47.435786, 47.5, 47.562858, 47.628064, 47.7,
            47.783047, 47.881586, 48.
        ])
        tc_track.equal_timestep(time_step_h=1)

        self.assertEqual(tc_track.data[0].time.size, 223)
        self.assertAlmostEqual(tc_track.data[0].lon.values[0], 167.207761)
        self.assertAlmostEqual(tc_track.data[0].lon.values[-1], -172)
        self.assertAlmostEqual(tc_track.data[0].lon.values[137], 179.75187272)
        self.assertAlmostEqual(tc_track.data[0].lon.values[138], 179.885288)
        self.assertAlmostEqual(tc_track.data[0].lon.values[139], -179.98060885)
        self.assertAlmostEqual(tc_track.data[0].lon.values[140], -179.84595743)
        self.assertAlmostEqual(tc_track.data[0].lat.values[0], 40.196053)
        self.assertAlmostEqual(tc_track.data[0].lat.values[-1], 48.)
        self.assertEqual(tc_track.data[0].time_step[7], 1)
        self.assertEqual(np.max(tc_track.data[0].radius_max_wind), 0)
        self.assertEqual(np.min(tc_track.data[0].radius_max_wind), 0)
        self.assertEqual(tc_track.data[0].max_sustained_wind[21], 25)
        self.assertAlmostEqual(tc_track.data[0].central_pressure.values[29], 1008, places=0)
        self.assertEqual(np.max(tc_track.data[0].environmental_pressure), 1010)
        self.assertEqual(np.min(tc_track.data[0].environmental_pressure), 1010)
        self.assertEqual(tc_track.data[0]['time.year'][13], 1951)
        self.assertEqual(tc_track.data[0]['time.month'][26], 8)
        self.assertEqual(tc_track.data[0]['time.day'][7], 27)
        self.assertEqual(tc_track.data[0].max_sustained_wind_unit, 'kn')
        self.assertEqual(tc_track.data[0].central_pressure_unit, 'mb')
        self.assertEqual(tc_track.data[0].orig_event_flag, 1)
        self.assertEqual(tc_track.data[0].name, '1951239N12334')
        self.assertEqual(tc_track.data[0].data_provider, 'hurdat_atl')
        np.testing.assert_array_equal(tc_track.data[0].basin, 'NA')
        self.assertEqual(tc_track.data[0].id_no, 1951239012334)
        self.assertEqual(tc_track.data[0].category, 1)

    def test_interp_origin_inv_pass(self):
        """Interpolate track to min_time_step crossing lat origin"""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK)
        tc_track.data[0].lon.values = np.array([
            167.207761, 168.1, 168.936535, 169.728947, 170.5, 171.257176,
            171.946822, 172.5, 172.871797, 173.113396, 173.3, 173.496375,
            173.725522, 174., 174.331591, 174.728961, 175.2, 175.747632,
            176.354929, 177., 177.66677, 178.362433, 179.1, 179.885288,
            -179.304661, -178.5, -177.726442, -176.991938, -176.3, -175.653595,
            -175.053513, -174.5, -173.992511, -173.527342, -173.1, -172.705991,
            -172.340823, -172.
        ])
        tc_track.data[0].lon.values = - tc_track.data[0].lon.values
        tc_track.data[0].lat.values = np.array([
            40.196053, 40.6, 40.930215, 41.215674, 41.5, 41.816354, 42.156065,
            42.5, 42.833998, 43.16377, 43.5, 43.847656, 44.188854, 44.5,
            44.764269, 44.991925, 45.2, 45.402675, 45.602707, 45.8, 45.995402,
            46.193543, 46.4, 46.615718, 46.82312, 47., 47.130616, 47.225088,
            47.3, 47.369224, 47.435786, 47.5, 47.562858, 47.628064, 47.7,
            47.783047, 47.881586, 48.
        ])
        tc_track.equal_timestep(time_step_h=1)

        self.assertEqual(tc_track.data[0].time.size, 223)
        self.assertAlmostEqual(tc_track.data[0].lon.values[0], -167.207761)
        self.assertAlmostEqual(tc_track.data[0].lon.values[-1], 172)
        self.assertAlmostEqual(tc_track.data[0].lon.values[137], -179.75187272)
        self.assertAlmostEqual(tc_track.data[0].lon.values[138], -179.885288)
        self.assertAlmostEqual(tc_track.data[0].lon.values[139], 179.98060885)
        self.assertAlmostEqual(tc_track.data[0].lon.values[140], 179.84595743)
        self.assertAlmostEqual(tc_track.data[0].lat.values[0], 40.196053)
        self.assertAlmostEqual(tc_track.data[0].lat.values[-1], 48.)
        self.assertEqual(tc_track.data[0].time_step[7], 1)
        self.assertEqual(np.max(tc_track.data[0].radius_max_wind), 0)
        self.assertEqual(np.min(tc_track.data[0].radius_max_wind), 0)
        self.assertEqual(tc_track.data[0].max_sustained_wind[21], 25)
        self.assertAlmostEqual(tc_track.data[0].central_pressure.values[29], 1008, places=0)
        self.assertEqual(np.max(tc_track.data[0].environmental_pressure), 1010)
        self.assertEqual(np.min(tc_track.data[0].environmental_pressure), 1010)
        self.assertEqual(tc_track.data[0]['time.year'][13], 1951)
        self.assertEqual(tc_track.data[0]['time.month'][26], 8)
        self.assertEqual(tc_track.data[0]['time.day'][7], 27)
        self.assertEqual(tc_track.data[0].max_sustained_wind_unit, 'kn')
        self.assertEqual(tc_track.data[0].central_pressure_unit, 'mb')
        self.assertEqual(tc_track.data[0].orig_event_flag, 1)
        self.assertEqual(tc_track.data[0].name, '1951239N12334')
        self.assertEqual(tc_track.data[0].data_provider, 'hurdat_atl')
        np.testing.assert_array_equal(tc_track.data[0].basin, 'NA')
        self.assertEqual(tc_track.data[0].id_no, 1951239012334)
        self.assertEqual(tc_track.data[0].category, 1)

    def test_dist_since_lf_pass(self):
        """Test _dist_since_lf for andrew tropical cyclone."""
        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TC_ANDREW_FL)
        track = tc_track.get_track()
        track['on_land'] = ('time', u_coord.coord_on_land(track.lat.values, track.lon.values))
        track['dist_since_lf'] = ('time', tc._dist_since_lf(track))

        msk = ~track.on_land
        self.assertTrue(np.all(np.isnan(track.dist_since_lf.values[msk])))
        self.assertEqual(track.dist_since_lf.values[msk].size, 38)

        self.assertGreater(
            track.dist_since_lf.values[-1],
            u_coord.dist_to_coast(track.lat.values[-1], track.lon.values[-1]) / 1000)
        self.assertEqual(1020.5431562223974, track['dist_since_lf'].values[-1])

        # check distances on land always increase, in second landfall
        dist_on_land = track.dist_since_lf.values[track.on_land]
        self.assertTrue(np.all(np.diff(dist_on_land)[1:] > 0))

    def test_category_pass(self):
        """Test category computation."""
        max_sus_wind = np.array([25, 30, 35, 40, 45, 45, 45, 45, 35, 25])
        max_sus_wind_unit = 'kn'
        cat = tc.set_category(max_sus_wind, max_sus_wind_unit)
        self.assertEqual(0, cat)

        max_sus_wind = np.array([25, 25, 25, 30, 30, 30, 30, 30, 25, 25, 20])
        max_sus_wind_unit = 'kn'
        cat = tc.set_category(max_sus_wind, max_sus_wind_unit)
        self.assertEqual(-1, cat)

        max_sus_wind = np.array([80, 90, 100, 115, 120, 125, 130,
                                 120, 110, 80, 75, 80, 65])
        max_sus_wind_unit = 'kn'
        cat = tc.set_category(max_sus_wind, max_sus_wind_unit)
        self.assertEqual(4, cat)

        max_sus_wind = np.array([
            28.769475, 34.52337, 40.277265, 46.03116, 51.785055, 51.785055,
            51.785055, 51.785055, 40.277265, 28.769475
        ])
        max_sus_wind_unit = 'mph'
        cat = tc.set_category(max_sus_wind, max_sus_wind_unit)
        self.assertEqual(0, cat)

        max_sus_wind = np.array([
            12.86111437, 12.86111437, 12.86111437, 15.43333724, 15.43333724,
            15.43333724, 15.43333724, 15.43333724, 12.86111437, 12.86111437,
            10.2888915
        ])
        max_sus_wind_unit = 'm/s'
        cat = tc.set_category(max_sus_wind, max_sus_wind_unit)
        self.assertEqual(-1, cat)

        max_sus_wind = np.array([
            148.16, 166.68, 185.2, 212.98, 222.24, 231.5, 240.76, 222.24,
            203.72, 148.16, 138.9, 148.16, 120.38
        ])
        max_sus_wind_unit = 'km/h'
        cat = tc.set_category(max_sus_wind, max_sus_wind_unit)
        self.assertEqual(4, cat)

    def test_estimate_params_pass(self):
        """Test track parameter estimation functions."""
        cen_pres = np.array([-999, 993, np.nan, -1, 0, 1004, np.nan])
        v_max = np.array([45, np.nan, 50, 55, 0, 60, 75])
        lat = np.array([13.8, 13.9, 14, 14.1, 14.1, np.nan, -999])
        lon = np.array([np.nan, -52.8, -54.4, -56, -58.4, -59.7, -61.1])
        ref_pres = np.array([np.nan, 993, 990, 986, np.nan, 1004, np.nan])
        out_pres = tc._estimate_pressure(cen_pres, lat, lon, v_max)
        np.testing.assert_array_almost_equal(ref_pres, out_pres, decimal=0)

        v_max = np.array([45, np.nan, 50, 55, 0, 60, 75])
        cen_pres = np.array([-999, 993, np.nan, -1, 0, 1004, np.nan])
        lat = np.array([13.8, 13.9, 14, 14.1, 14.1, np.nan, -999])
        lon = np.array([np.nan, -52.8, -54.4, -56, -58.4, -59.7, -61.1])
        ref_vmax = np.array([45, 46, 50, 55, np.nan, 60, 75])
        out_vmax = tc._estimate_vmax(v_max, lat, lon, cen_pres)
        np.testing.assert_array_almost_equal(ref_vmax, out_vmax, decimal=0)

        roci = np.array([np.nan, -1, 145, 170, 180, 0, -5])
        cen_pres = np.array([-999, 993, np.nan, -1, 0, 1004, np.nan])
        ref_roci = np.array([np.nan, 182.792715, 145, 170, 180, 161.5231086, np.nan])
        out_roci = tc.estimate_roci(roci, cen_pres)
        np.testing.assert_array_almost_equal(ref_roci, out_roci)

        rmw = np.array([17, 33, -1, 25, np.nan, -5, 13])
        cen_pres = np.array([-999, 993, np.nan, -1, 0, 1004, np.nan])
        ref_rmw = np.array([17, 33, np.nan, 25, np.nan, 43.95543761, 13])
        out_rmw = tc.estimate_rmw(rmw, cen_pres)
        np.testing.assert_array_almost_equal(ref_rmw, out_rmw)

    def test_estimate_rmw_pass(self):
        """Test estimate_rmw function."""
        NM_TO_KM = (1.0 * ureg.nautical_mile).to(ureg.kilometer).magnitude

        tc_track = tc.TCTracks.from_processed_ibtracs_csv(TEST_TRACK)
        tc_track.equal_timestep()
        rad_max_wind = tc.estimate_rmw(
            tc_track.data[0].radius_max_wind.values,
            tc_track.data[0].central_pressure.values) * NM_TO_KM

        self.assertAlmostEqual(rad_max_wind[0], 87, places=0)
        self.assertAlmostEqual(rad_max_wind[10], 87, places=0)
        self.assertAlmostEqual(rad_max_wind[128], 56, places=0)
        self.assertAlmostEqual(rad_max_wind[129], 55, places=0)
        self.assertAlmostEqual(rad_max_wind[130], 54, places=0)
        self.assertAlmostEqual(rad_max_wind[189], 53, places=0)
        self.assertAlmostEqual(rad_max_wind[190], 55, places=0)
        self.assertAlmostEqual(rad_max_wind[191], 57, places=0)
        self.assertAlmostEqual(rad_max_wind[192], 58, places=0)
        self.assertAlmostEqual(rad_max_wind[200], 71, places=0)

    def test_tracks_in_exp_pass(self):
        """Check if tracks in exp are filtered correctly"""

        # Load two tracks from ibtracks
        storms = {'in': '2000233N12316', 'out': '2000160N21267'}
        tc_track = tc.TCTracks.from_ibtracs_netcdf(storm_id=list(storms.values()))

        # Define exposure from geopandas
        world = gpd.read_file(gpd.datasets.get_path('naturalearth_lowres'))
        exp_world = Exposures(world)
        exp = Exposures(exp_world.gdf[exp_world.gdf.name=='Cuba'])

        # Compute tracks in exp
        tracks_in_exp = tc_track.tracks_in_exp(exp, buffer=1.0)

        self.assertTrue(tracks_in_exp.get_track(storms['in']))
        self.assertFalse(tracks_in_exp.get_track(storms['out']))

    def test_get_landfall_idx(self):
        """Test identification of landfalls"""
        tr_ds = xr.Dataset()
        datetimes = list()
        for h in range(0, 24, 3):
            datetimes.append(dt(2000, 1, 1, h))
        tr_ds.coords['time'] = ('time', datetimes)
        # no landfall
        tr_ds['on_land'] = np.repeat(np.array([False]), 8)
        sea_land_idx, land_sea_idx = tc._get_landfall_idx(tr_ds)
        self.assertEqual([len(sea_land_idx), len(land_sea_idx)], [0,0])
        # single landfall
        tr_ds['on_land'] = np.array([False, False, True, True, True, False, False, False])
        sea_land_idx, land_sea_idx = tc._get_landfall_idx(tr_ds)
        self.assertEqual([len(sea_land_idx), len(land_sea_idx)], [1,1])
        self.assertEqual([sea_land_idx, land_sea_idx], [2, 5])
        # single landfall from starting point
        tr_ds['on_land'] = np.array([True, True, True, True, True, False, False, False])
        sea_land_idx, land_sea_idx = tc._get_landfall_idx(tr_ds)
        self.assertEqual([len(sea_land_idx), len(land_sea_idx)], [0,0])
        sea_land_idx, land_sea_idx = tc._get_landfall_idx(tr_ds, include_starting_landfall=True)
        self.assertEqual([sea_land_idx, land_sea_idx], [0, 5])
        # two landfalls
        tr_ds['on_land'] = np.array([False, True, True, False, False, False, True, True])
        sea_land_idx, land_sea_idx = tc._get_landfall_idx(tr_ds)
        self.assertEqual([len(sea_land_idx), len(land_sea_idx)], [2,2])
        self.assertEqual(sea_land_idx.tolist(), [1,6])
        self.assertEqual(land_sea_idx.tolist(), [3,8])
        # two landfalls, starting on land
        tr_ds['on_land'] = np.array([True, True, False, False, True, True, False, False])
        sea_land_idx, land_sea_idx = tc._get_landfall_idx(tr_ds)
        self.assertEqual([sea_land_idx, land_sea_idx], [4, 6])
        sea_land_idx, land_sea_idx = tc._get_landfall_idx(tr_ds, include_starting_landfall=True)
        self.assertEqual(sea_land_idx.tolist(), [0,4])
        self.assertEqual(land_sea_idx.tolist(), [2,6])


# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestFuncs)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestIO))
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestIbtracs))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test Tag class
"""

import unittest

from climada.entity.tag import Tag

class TestAppend(unittest.TestCase):
    """Test loading funcions from the Hazard class"""

    def test_append_different_increase(self):
        """Appends an other tag correctly."""
        tag1 = Tag('file_name1.mat', 'dummy file 1')
        self.assertEqual('file_name1.mat', tag1.file_name)
        self.assertEqual('dummy file 1', tag1.description)

        tag2 = Tag('file_name2.mat', 'dummy file 2')

        tag1.append(tag2)

        self.assertEqual('file_name1.mat + file_name2.mat', tag1.file_name)
        self.assertEqual('dummy file 1 + dummy file 2', tag1.description)

    def test_append_equal_same(self):
        """Appends an other tag correctly."""
        tag1 = Tag('file_name1.mat', 'dummy file 1')
        tag2 = Tag('file_name1.mat', 'dummy file 1')

        tag1.append(tag2)
        self.assertEqual('file_name1.mat', tag1.file_name)
        self.assertEqual('dummy file 1', tag1.description)

    def test_append_empty(self):
        """Appends an other tag correctly."""
        tag1 = Tag('file_name1.mat', 'dummy file 1')
        tag2 = Tag()

        tag1.append(tag2)
        self.assertEqual('file_name1.mat', tag1.file_name)
        self.assertEqual('dummy file 1', tag1.description)

        tag1 = Tag()
        tag2 = Tag('file_name1.mat', 'dummy file 1')

        tag1.append(tag2)
        self.assertEqual('file_name1.mat', tag1.file_name)
        self.assertEqual('dummy file 1', tag1.description)

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestAppend)
    unittest.TextTestRunner(verbosity=2).run(TESTS)
